# Options.py
# $Id: Options.py,v 1.18 2001/07/14 23:26:27 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# Eventually this will hold all program options, and maybe be incorporated
# into some sort of options editor.  (Hopefully)
#
# For now, it holds information about what the program should and shouldn't
# do.  Variable names should be pretty self explanatory.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#############################################################################
import os
from string import *
import Cache
import Associations

class Options:
    def __init__(self, *args):
        # Default values for some important options...
        self.ip_cache = {}
        self.cache = Cache.Cache()
        self.associations = Associations.Associations()
        self.setDefaultOpts()
        self.greenLight()
        return None

    # Accessors/Mutators
    
    def getCache(self):
        return self.cache
    
    def setCache(self, newcache):
        self.cache = newcache
        return self.getCache()
    
    def getAssociations(self):
        return self.associations
    
    def setAssociations(self, newassoc):
        self.associations = newassoc
        return self.getAssociations()

    def getIP(self, hostname):
        """Return the cached IP of hostname.  May throw KeyError"""
        return self.ip_cache[hostname]
    
    def setIP(self, hostname, IP):
        self.ip_cache[hostname] = IP
        return self.getIP(hostname)

    def save(self, alternate_filename=None):
        """Saves all options to the prefs directory/forgrc unless an
        alternate filename is specified.  Throws all IOErrors out"""
        if not alternate_filename:
            prefs = self.getOption('prefs_directory')
            filename = "%s%s%s" % (prefs, os.sep, "forgrc")
        else:
            filename = alternate_filename

        fp = open(filename, "w")
        fp.write(self.toString())
        fp.flush()
        fp.close()
        return 1
    
    def toggle(self, key):
        """Flip the value of the option specified by key.  If it was true,
        it becomes false, and if it was false, it becomes true."""
        try:
            val = self.opts[key]

            if val:
                print "Toggle(%s): FALSE" % key
                self.opts[key] = None
            else:
                print "Toggle(%s): TRUE" % key
                self.opts[key] = 1
        except:
            print "Toggle(%s): TRUE" % key
            self.opts[key] = 1
        return self.opts[key]
            
    def greenLight(self):
        """Set the green light.  This is used for thread synchronization"""
        self.GREEN_LIGHT = 1
        return self.GREEN_LIGHT
    
    def redLight(self):
        """Turn off the green light.  For thread synchronization."""
        self.GREEN_LIGHT = None
        return self.GREEN_LIGHT
    
    def exists(self, filename):
        """Returns true if filename exists, false otherwise."""
        try:
            stuff = os.stat(filename)
            return 1
        except OSError:
            return None
        return None
            
    def setDefaultOpts(self):
        """Sets default set of options so that the structure is not empty."""
        self.opts = {}

        if os.sep == '\\':
            # Ugly hack for windows.  Some versions of windows
            # yield c:\ for the home directory, and some return something
            # really screwed up like \ or c\ even.
            self.opts['prefs_directory'] = "C:\\FORG-DATA"
        else:
            self.opts['prefs_directory'] = "%s%s%s" % (os.path.expanduser("~"),
                                                       os.sep, ".forg")

        self.opts['home'] = "gopher://gopher.floodgap.com:70/1/"
            
        if not self.exists(self.opts['prefs_directory']):
            os.mkdir(self.opts['prefs_directory'])

        self.opts['use_cache']                   = 1
        self.opts['delete_cache_on_exit']        = None
        self.opts['use_url_format']              = 1
        self.opts['grab_resource_info']          = None
        self.opts['show_host_port']              = None
        self.opts['save_options_on_exit']        = 1
        self.opts['strip_carraige_returns']      = 1
        self.opts['show_cached']                 = None
        self.opts['display_info_in_directories'] = None
        self.opts['use_PIL']                     = 1  # Use PIL for images

        basedir = self.getOption('prefs_directory')

        pkey = 'prefs_directory'

        self.opts['cache_directory'] = "%s%s%s" % (self.opts[pkey],
                                                   os.sep,
                                                   "cache")

        if not self.exists(self.opts['cache_directory']):
            os.mkdir(self.opts['cache_directory'])
        
        self.opts['cache_prefix'] = "%s%s" % (self.opts['cache_directory'],
                                              os.sep)
        return None

    def makeToggleWrapper(self, keyname):
        """Returns a function which when called with no arguments toggles the
        value of keyname within the options structure.  This is used for menu
        callbacks connected to check buttons."""

        def toggle_wrapper(opts=self, key=keyname):
            return opts.toggle(key)
        
        return toggle_wrapper

    def setOption(self, optionname, optionvalue):
        """Set optionname to optionvalue"""
        self.opts[optionname] = optionvalue
        return self.getOption(optionname)
        
    def getOption(self, optionname):
        """Get an option named optionname."""
        try:
            optionname = lower(optionname)
            return self.opts[optionname]
        except KeyError:
            return None
        
    def toString(self):
        """Returns string representation of the object."""
        return self.__str__()

    def __repr__(self):
        return self.__str__()
    
    def __str__(self):
        # God I love the map() function.
        lines = map(lambda x, self=self: "%s = %s" % (x, self.opts[x]),
                    self.opts.keys())
        comments = "%s%s%s" % ("# Options for the FORG\n",
                               "# Please don't edit me unless you know what\n",
                               "# you're doing.\n")
        return comments + join(lines, "\n") + "\n"
    
    def parseFile(self, filename):
        """Parse filename into a set of options.  Caller is responsible
        for catching IOError related to reading files."""
        print "Previously had %d keys" % len(self.opts.keys())
        self.setDefaultOpts()
        fp = open(filename, "r")

        line = fp.readline()
        line_num = 0
        while line != '':
            line_num = line_num + 1
            commentIndex = find(line, "#")
            if commentIndex != -1:
                line = line[0:commentIndex]

            line = strip(line)
            
            if line == '':             # Nothing to grokk
                line = fp.readline()   # Get next line...
                continue
            
            items = split(line, "=")
            if len(items) < 2:
                print "Options::parseFile: no '=' on line number %d" % line_num
                line = fp.readline()   # Get next line...
                continue
            if len(items) > 2:
                print("Options::parseFile: too many '=' on line number %d" %
                      line_num)
                line = fp.readline()   # Get next line...
                continue
            
            key = lower(strip(items[0]))       # Normalize and lowercase
            val = lower(strip(items[1]))

            # Figure out what the hell val should be
            if val == 'no' or val == 'none' or val == 0 or val == '0':
                val = None
                
            self.opts[key] = val
            line = fp.readline()

        return self

# Here is one instance of an options structure.
# This is used by TkGui and other modules as a centralized place
# to store program options.
program_options = Options()




# GUIQuestion.py
# $Id: GUIQuestion.py,v 1.13 2001/07/02 22:50:39 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
#
# A GUI representation of a question.  It should provide all methods needed
# to get the response, set default values, etc.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#############################################################################

from Tkinter import *
import Pmw
import tkFileDialog
from string import *
from gopher import *
import GopherResource
import GopherResponse
import GopherConnection
import ContentFrame
import Cache
import Options
import Question

GUIQuestionException = "Sorry.  Things just sometimes go wrong."

class GUIQuestion(Frame):
    verbose = None

    def __init__(self, parent_widget, question):
        Frame.__init__(self, parent_widget)  # Superclass constructor
        self.parent = parent_widget
        self.question = question
        self.type     = self.question.getType()

        promptString = self.question.getPromptString()
        defaultValue = self.question.getDefault()

        Label(self, text=promptString).grid(row=0, col=0, sticky=W)        

        if self.type == QUESTION_NOTE:
            # Prompt string is all we need for this one.
            return None

        if self.type == QUESTION_ASK:
            self.entry = Entry(self)
            if len(defaultValue) > 0:
                self.entry.insert('end', defaultValue)
            self.entry.grid(row=0, col=1, columnspan=4, sticky=W)
            return None
        if self.type == QUESTION_ASKF or self.type == QUESTION_CHOOSEF:
            self.entry = Entry(self)
            if len(defaultValue) > 0:
                self.entry.insert('end', defaultValue)
            self.entry.grid(row=0, col=1, columnspan=4, sticky=W)

            # Browse buttons for file selection.
            self.browse = Button(text="Browse", command=self.browse)
            self.browse.grid(row=0, col=5, sticky=W)
            return None
        if self.type == QUESTION_ASKP:
            self.entry = Entry(self, show="*")

            if len(defaultValue) > 0:
                self.entry.insert('end', defaultValue)
            self.entry.grid(row=0, col=1, columnspan=4, sticky=W)
            return None
        if self.type == QUESTION_ASKL:
            self.entry = Pmw.ScrolledText(self, hscrollmode='dynamic',
                                          text_width=80, text_height=6,
                                          vscrollmode='dynamic')
            self.entry.grid(row=1, col=0, columnspan=2, rowspan=2,
                            sticky='N')
            return None
        if self.type == QUESTION_SELECT:
            self.entry = Pmw.RadioSelect(self, buttontype='checkbutton',
                                         command=self.changed)
            
            for opt in self.question.options:
                self.entry.add(opt)
                
            if defaultValue:
                print "Invoking defalut %s" % defaultValue
                self.entry.invoke(defaultValue)

            self.entry.grid(row=1, col=0, columnspan=4, rowspan=4,
                            sticky='NSEW')
            print 'Returning SELECT GUIQuestion'
            return None
        if self.type == QUESTION_CHOOSE:
            self.entry = Pmw.RadioSelect(self, buttontype='radiobutton',
                                         command=self.changed)
            for opt in self.question.options:
                self.entry.add(opt)
            if defaultValue:
                print "Invoking defalut %s" % defaultValue
                self.entry.invoke(defaultValue)
                
            self.entry.grid(row=1, col=0, columnspan=4, rowspan=4,
                            sticky='NSEW')
            print "Returning CHOOSE GUIQuestion"
            return None
        return None
    def browse(self, *args):
        dir = os.path.abspath(os.getcwd())
        filename = tkFileDialog.asksaveasfilename(initialdir=dir)
        self.entry.delete(0, 'end')
        if filename:
            self.entry.insert('end', filename)
        return None
    def getType(self):
        return self.question.getType()
    def changed(self, *args):
        print "Selection changed:  Current selection:  ", args
    def getResponse(self):
        """Returns the current state of the widget, or what should be sent
        to the server."""
        if self.entry.__class__ == Entry:
            return "%s\n" % self.entry.get()
        elif self.entry.__class__ == Pmw.ScrolledText:
            buf = self.entry.get()
            lines = count(buf, "\n")
            return "%d\n%s\n" % (lines, buf)
        elif self.entry.__class__ == Pmw.RadioSelect:
            list = self.entry.getcurselection()
            return "%s\n" % list[0]
        else:
            # Huh?  What?  Eh?  WTF is going on?
            raise GUIQuestionException, "Cannot get content: Unknown type"
        return ""

        



# TkGui.py
# $Id: TkGui.py,v 1.56 2001/07/14 23:50:49 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
#
# This handles the GUI of the program, displaying, bookmarks, etc.
# This module is usually called by the starter script.  All other modules
# are usually invoked through this one as the GUI needs them.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
############################################################################
# Python-wide modules not specific to this program
import os
import string
import sys
import re
from threading import *
import time
import pickle
import socket   # Just for errors 

# Data structures, connections, and the like.
import utils
import Options
import GopherObject
import ResourceInformation
import GopherResource
import GopherResponse
import GopherConnection
import Bookmark
import Associations
import Cache
from gopher import *

# GUI specific modules.
from Tkinter import *
import tkFileDialog               # File save/open dialog
import AssociationsEditor         # GUI editor for association rules
import forg                       # The guts of the entire application
import BookmarkEditor             # Editing bookmarks window
import Pmw                        # Python Mega Widgets
import Dialogs                    # FORG specific dialogs

ContentException = "ContentException:"

class TkGui(Tk):
    verbose = None
    MENU_FILE       = 1
    MENU_EDIT       = 2
    MENU_NAVIGATION = 3
    MENU_BOOKMARKS  = 4
    MENU_OPTIONS    = 5
    MENU_HELP       = 6

    RED             = '#FF0000'
    GREEN           = '#00FF00'
    BLUE            = '#0000FF'
    
    def __init__(self, URL=None):
        """Create an instance of the gopher program."""        
        Tk.__init__(self)
        
        self.loadOptions()              # Load program options from disk
        self.loadBookmarks()            # Load program bookmarks from disk
        self.createAssociations()       # Load program associations

        self.options_filename = self.getPrefsDirectory() + os.sep + "options"

        # Load options into the root window and all subwindows.
        try:
            if utils.file_exists(self.options_filename):
                self.option_readfile(self.options_filename)
        except Exception, errstr:
            print("***Error loading options from %s: %s" %
                  (self.options_filename, errstr))

        if self.verbose:
            print "OPTIONS:\n%s" % Options.program_options.toString()

        self.minsize(550, 550)
        self.title("FORG v. %s" % forg.getVersion())

        self.saveFileDialog        = None
        self.DOWNLOAD_ME           = None     # Thread control variable
        self.CURRENTLY_DOWNLOADING = None     # Are we downloading right now?
        self.LAUNCH_ME             = None     # Thread control variable
        self.WIDGET_ENTRY          = 0

        self.mainBox = Frame(self)
        self.mainBox.pack(fill='both', expand=1)
        self.make_menus()
        self.config(menu=self.menu)
        self.buttonBar = Frame(self.mainBox)
        self.navBox    = Frame(self.mainBox)
        self.buttonBar.pack(fill='x', expand=0)
        self.navBox.pack(fill='x', expand=0)

        self.BACKWARD = Button(self.buttonBar, text='Back',
                               command=self.goBackward)
        self.FORWARD  = Button(self.buttonBar, text='Forward',
                               command=self.goForward)
        # self.STOP   = Button(self.buttonBar, text='Stop',
        #                      command=self.stop)
        self.RELOAD   = Button(self.buttonBar, text='Reload',
                               command=self.reload)
        self.HOME     = Button(self.buttonBar, text='Home',
                               command=self.goHome)
        
        self.BACKWARD.pack(side='left')
        self.FORWARD.pack(side='left')
        # self.STOP.pack(side='left')
        self.RELOAD.pack(side='left')
        self.HOME.pack(side='left')

        if Options.program_options.getOption('use_url_format'):
            self.urlLabel = Label(self.navBox, text="Location:")
            self.urlEntry = Entry(self.navBox)
            self.urlEntry.bind("<Return>", self.go)
            
            self.urlLabel.pack(side='left')
            self.urlEntry.pack(side='left', expand=1, fill='x')
        else:
            self.hostLabel = Label(self.navBox, text='Host:')
            self.hostEntry = Entry(self.navBox)
            self.resourceLabel = Label(self.navBox, text='Resource:')
            self.resourceEntry = Entry(self.navBox, text="/")
            self.resourceEntry.insert('end', '/')
            self.portLabel = Label(self.navBox, text='Port:')
            self.portEntry = Entry(self.navBox, width=5)

            self.portEntry.insert('end', '70')
            self.hostLabel.grid(row=0, col=0, sticky=W)
            self.hostEntry.grid(row=0, col=1, columnspan=2, sticky=W)
            self.resourceLabel.grid(row=0, col=3, sticky=W)
            self.resourceEntry.grid(row=0, col=4, columnspan=2, sticky=W)
            self.portLabel.grid(row=0, col=6, sticky=W)
            self.portEntry.grid(row=0, col=7, sticky=W) # No colspan: short box

        self.gobutton = Button(self.navBox, text='Go', command=self.go)

        if Options.program_options.getOption('use_url_format'):
            self.gobutton.pack(side='right')
        else:
            self.gobutton.grid(row=0, col=8, sticky=W)

        if URL != None:
            resource = GopherResource.GopherResource()
            resource.setURL(URL)
        else:
            resource = GopherResource.GopherResource()
            resource.setURL("gopher://gopher.quux.org/")
            resource.setTitle("QUUX.org Root")

        self.CONTENT_BOX = forg.FORG(parent_widget=self.mainBox,
                                     parent_object=self,
                                     resource=resource)
        
        self.messageBar = Pmw.MessageBar(self.mainBox,
                                         entry_width = 80,
                                         entry_relief='groove',
                                         labelpos = 'w',
                                         label_text = 'Status:')

        self.CONTENT_BOX.setMessageBar(self.messageBar)
        
        self.CONTENT_BOX.pack(expand=1, fill='both')
        self.messageBar.pack(expand=0, fill='x')
        utils.msg(self.messageBar, "Ready")
        
        # Call fn when the window is destroyed.
        self.protocol("WM_DELETE_WINDOW", self.destroy)
        self.setLocation(self.CONTENT_BOX.getResource())
        return None

    def getCache(self):
        return Options.program_options.cache

    def stats(self, *args):
        return self.CONTENT_BOX.stats()

    def destroy(self, *args):
        """Overridden destroy method for the application.  This does
        cleanup, particularly with the various living threads before
        destroying the GUI on screen"""
        
        self.DOWNLOAD_ME = "DIE"    # Kill the two other threads with this
        self.LAUNCH_ME   = "DIE"    # 'message' to them.
                                    # They will see this.

        # Things that need to be done before we exit...

        if Options.program_options.getOption('save_options_on_exit'):
            # self.messageBar.message('state', "Saving program options...")
            print "Saving options on exit..."
            self.saveOptions()

        if Options.program_options.getOption('delete_cache_on_exit'):
            print "Deleting cache on exit..."
            Options.program_options.cache.deleteCacheNoPrompt()
        
        Tk.destroy(self)
        print "MT:  Exit"

    def editAssociations(self, *args):
        """Callback: opens the associations editor and lets the users change
        which programs launch on which filetypes."""
        x = AssociationsEditor.AssociationsEditor(
            self, Options.program_options.associations
            )
        return None

    def notYetImplemented(self, *args):
        """Opens a dialog for the user to see that the feature hasn't been
        implemented yet."""
        self.genericMessage(
            "Contribute to the Programmer Pizza Fund today!\n" +
            "Maybe that will help get this feature put into\n" +
            "the program!")
        return None
    
    def setAssociations(self, assocList):
        """Modifier: associates a list of associations usually from the
        AssociationsEditor to this object."""
        Options.program_options.setAssociations(assocList)
        return None
    
    def createAssociations(self, *args):
        """Creates a set of default associations.  These are mostly UNIX
        centric unless I change them and forget to update this docstring.  :)
        """
        Options.program_options.setAssociations(Associations.Associations())

        filename = self.getPrefsDirectory() + os.sep + "forg-associations"
        self.loadAssociations(filename)

        if Options.program_options.getAssociations().isEmpty():
            print "ADDING DEFAULT ASSOCIATIONS"
            # Add defaults if there isn't anything in the association list.
            images = [".gif", ".jpg", ".bmp", ".xpm", ".xbm",
                      ".png", ".jpeg", ".tif", ".tiff" ]

            for item in images:
                Options.program_options.associations.addAssociation(item,
                                                                    "eeyes $1")

            browser_stuff = [".html", ".htm", ".css"]
            for item in browser_stuff:
                cmd = "netscape $1"
                Options.program_options.associations.addAssociation(item, cmd)

            Options.program_options.associations.addAssociation(".pdf",
                                                                "xpdf $1")
            Options.program_options.associations.addAssociation(".ps",
                                                                "gv $1")
        return None

    def genericMessage(self, str, title='Information:'):
        """Given a string, pops up a dialog with the appropriate title and
        content string."""
        x = Dialogs.InformationDialog(self, str, title)
        return None
        
    def genericError(self, errstr, title='Error:'):
        """Given an error string, pops up a dialog with the same title and
        text, alerting the user to something or other."""
        x = Dialogs.ErrorDialog(self, errstr, title)
        return None

    def find(self, *args):
        """Wrapper function for find.  When the find feature is used, this
        gets called to call the find feature in the child widget."""
        
        wid = self.CONTENT_BOX.navList.getCurrent().getData().getWidget()
        self._find_dialog = Dialogs.FindDialog(self, wid, self)
        return None

    def addBookmark(self, *args):
        """Callback: take the current content frame and add the resource
        associated to it to the bookmark list.  The new bookmark is appended
        at the end of the bookmarks menu"""

        resource = self.CONTENT_BOX.getResource()
        foo = Bookmark.Bookmark(resource)

        # Save the bookmark...
        self.bookmarks.insert(Bookmark.BookmarkMenuNode(foo))
        self.saveBookmarks()

        self.reloadBookmarks()
        return None

    def reloadBookmarks(self, *args):
        """Reloads bookmarks from disk, and replaces the current Bookmark
        menu with the new one loaded from disk."""
        
        print "*** Reloading bookmarks from disk after edit."
        self.loadBookmarks()

        # self.bookmarks now holds the freshly edited copy of the bookmarks.
        # Create a new Tk menu based off of them.

        # Destroy the old menu and insert the new.
        self.bookmarkTkMenu = self.buildTkBookmarksMenu(self.menu,
                                                        self.bookmarks)

        self.menu.delete(self.MENU_BOOKMARKS)
        self.menu.insert_cascade(index=self.MENU_BOOKMARKS,
                                 label=self.bookmarks.getName(),
                                 menu=self.bookmarkTkMenu)
        return None

    def getPrefsDirectory(self):
        return Options.program_options.getOption('prefs_directory')

    def loadAssociations(self, filename=None):
        if filename is None:
            filename = self.getPrefsDirectory() + os.sep + "forg-associations"

        try:
            Options.program_options.associations.loadFromFile(filename)
        except IOError, errstr:
            print "****Cannot load associations from %s: %s" % (filename, str)
            return 0 # Failure
        return 1  # Success

    def loadOptions(self, filename=None):
        if filename is None:
            filename = self.getPrefsDirectory() + os.sep + "forgrc"

        try:
            Options.program_options.parseFile(filename)
        except IOError, errstr:
            print "**** Couldn't parse options at %s: %s" % (filename, errstr)
            return None
        
        print "****Successfully loaded options from disk."
        return 1
    
    def saveOptions(self, *args):
        """Saves the user options to a file in their home directory.  Who knows
        what happens on windows boxen."""

        filename = self.getPrefsDirectory() + os.sep + "forg-associations"

        try:
            Options.program_options.associations.writeToFile(filename)
        except IOError, str:
            print "***Error saving associations to disk: %s" % str

        ofilename = self.getPrefsDirectory() + os.sep + "forgrc"
        
        try:
            Options.program_options.save()
        except IOError, str:
            self.genericError("Couldn't write options to file:\n%s" % str)

        utils.msg(self.messageBar, "Finished saving options.")
        return None

    def loadBookmarks(self, *args):
        self.bookmarks = Bookmark.BookmarkMenu("Bookmarks")
        filename = self.getPrefsDirectory() + os.sep + "bookmarks"
        
        try:
            fp = open(filename, 'r')

            factory = Bookmark.BookmarkFactory()
            factory.parseResource(fp)
            self.bookmarks = factory.getMenu()
        except IOError, errstr:
            print "****Couldn't load bookmarks at %s: %s" % (filename, errstr)
            return None
        
        print "****Bookmarks successfully loaded from disk."
        return 1
    
    def saveBookmarks(self, *args):
        filename = self.getPrefsDirectory() + os.sep + "bookmarks"

        try:
            fp = open(filename, "w")

            factory = Bookmark.BookmarkFactory()
            factory.writeXML(fp, self.bookmarks)
            fp.close()
        except IOError, errstr:
            print "****Couldn't save bookmarks to %s: %s" % (filename, errstr)
            return None
        print "****Bookmarks successfully saved to %s" % filename
        return 1

    def stop(self, *args):
        """Just set green light to a false value, and wait for the download
        thread to get the point.  It will quit once it sees this."""
        return Options.program_options.redLight()

    def goForward(self, *args):
        self.CONTENT_BOX.goForward()
        self.setLocation(self.CONTENT_BOX.getResource())
    
    def goBackward(self, *args):
        self.CONTENT_BOX.goBackward()
        self.setLocation(self.CONTENT_BOX.getResource())
    
    def go(self, *rest):
        """This is what happens when the 'Go' button is clicked.  Information
        about the host, port, locator is fetched from the entry boxes, and
        the program goes there.  (Or tries, anyway)"""

        if Options.program_options.getOption('use_url_format'):
            url = string.strip(self.urlEntry.get())

            ind = string.find(url, "://")
            if ind != -1:
                proto = url[0:ind]
                if string.lower(proto) != "gopher":
                    self.genericError("Protocol\n\"%s\"\nnot supported." %
                                      proto)
                    return None
            elif string.lower(url[0:9]) != "gopher://":
                url = "gopher://" + url
                    
            res = GopherResource.GopherResource()
            try:
                res.setURL(url)
            except Exception, estr:
                self.genericError("Invalid gopher URL:\n%s\n%s" % url, estr)
                return None
        else:
            host    = self.hostEntry.get()
            port    = self.portEntry.get()
            locator = self.resourceEntry.get()
            
            if locator == '':
                locator = "/"

            if port == '' or port < 0:
                port = 70
        
            res  = GopherResource.GopherResource('1', host, port,
                                                 locator, "%s Root" % host)

        # Either way, go there.
        self.goElsewhere(res)

    def goElsewhere(self, resource, usecache=1, *args):
        self.CONTENT_BOX.goElsewhere(resource, usecache, args)
        self.setLocation(self.CONTENT_BOX.getResource())

    def openURL(self, URL):
        resource = GopherResource.GopherResource()
        resource.setURL(URL)
        return self.goElsewhere(resource)
        
    def showOpenURL(self, *args):
        d = Dialogs.OpenURLDialog(self, self.openURL)
        return None

    def saveFile(self, *args):
        return self.CONTENT_BOX.saveFile()

    def popupMenu(self, event):
        """Display pop-up menu on right click on a message"""
        self.popup.tk_popup(event.x_root, event.y_root)
        
    def createPopup(self):
        """Pop-up menu on right click on a message"""
        self.popup = Menu(self)
        self.popup['tearoff'] = FALSE
        self.popup.add_command(label='Save',
                               command=self.CONTENT_BOX.saveFile)
        self.popup.add_command(label='Back',
                               command=self.CONTENT_BOX.goBackward)
        self.popup.add_command(label='Forward',
                               command=self.CONTENT_BOX.goForward)

    def change_content_hook(self, resource):
        """This function is called by the child FORG instance whenever
        content is changed.  The resource argument is the GopherResource
        object corresponding to the content currently in the window."""
        return self.setLocation(resource)

    def setLocation(self, resource):
        """Takes a resource, and sets the location information at the top
        of the screen to the information in the resource.  If you're going to
        a certain location, this gets called to update what the user sees as
        the location."""

        if Options.program_options.getOption('use_url_format'):
            URL = resource.toURL()

            # Translate spaces.  This is ONLY done because this is the way
            # other programs like to get their URLs, not because it needs to
            # be this way per se.  Users copy and paste out of this location
            # bar, so put a standard style URL there.
            URL = re.sub(" ", "%20", URL)
            self.urlEntry.delete(0, 'end')
            self.urlEntry.insert('end', URL)
        else:
            self.hostEntry.delete(0, 'end')
            self.resourceEntry.delete(0, 'end')
            self.portEntry.delete(0, 'end')

            self.hostEntry.insert('end', resource.getHost())
            self.resourceEntry.insert('end', resource.getLocator())
            self.portEntry.insert('end', resource.getPort())

        return None

    def dumpQueue(self, *rest):
        return self.CONTENT_BOX.dumpQueue()

    def reload(self, *rest):
        return self.CONTENT_BOX.reload()

    def about(self, *args):
        return self.CONTENT_BOX.about()

    def quit(self, *args):
        """Quit the entire program.  Caller may have something after this."""
        self.destroy()

    def editBookmarks(self, *args):
        ed = BookmarkEditor.BookmarkEditor(self.bookmarks,
                                           ondestroy=self.reloadBookmarks)

    def buildTkBookmarksMenu(self, parent_menu, bookmarks):
        def fn(item, self=self):
            return self.goElsewhere(item)
        
        newTkMenu = bookmarks.getTkMenu(parent_menu, fn)
        newTkMenu.insert_separator(index=1)
        newTkMenu.insert_command(index=0, label="Edit Bookmarks",
                                 command=self.editBookmarks)
        newTkMenu.insert_command(index=1, label="Bookmark this page",
                                 command=self.addBookmark)
        return newTkMenu

    def goHome(self, *args):
        """Redirects the content of the main window to the user's chosen
        home site."""
        
        url = Options.program_options.getOption('home')

        if url:
            res = GopherResource.GopherResource()
            res.setURL(url)
            return self.goElsewhere(res)
        else:
            d = Dialogs.ErrorDialog(self,
                                    "You don't have a home site specified",
                                    "Error:  Cannot locate home site")
            return None
        
    def setHome(self, *args):
        url = self.CONTENT_BOX.getResource().toURL()
        Options.program_options.setOption('home', url)
        return None
    
    def make_menus(self):
        """Create the menuing system"""
        self.menu = Menu(self.mainBox)
        self.filemenu = Menu(self.menu)
        self.filemenu.add_command(label='Open URL', command=self.showOpenURL)
        self.filemenu.add_command(label='Save', command=self.saveFile)
        self.filemenu.add_command(label='Quit', 
                                  underline=0, command=self.quit)

        self.editmenu = Menu(self.menu)
        self.editmenu.add_command(label="Find", command=self.find)

        self.navmenu = Menu(self.menu)
        self.navmenu.add_command(label="Forward",
                                 command=self.goForward)
        self.navmenu.add_command(label="Backward",
                                 command=self.goBackward)
        self.navmenu.add_command(label="Reload",
                                 command=self.reload)

        self.optionsmenu = Menu(self.menu)
        self.optionsmenu.add_command(label='Associations',
                                     command=self.editAssociations)
        self.optionsmenu.add_command(label="Save Options/Associations",
                                     command=self.saveOptions)
        self.optionsmenu.add_command(label="Reload Options/Associations",
                                     command=self.loadOptions)
        self.optionsmenu.add_command(label="Set This Site as My Home",
                                     command=self.setHome)

        def purgeCacheWrapper(c=Options.program_options.cache, parent=self):
            c.emptyCache(parent)
            
        self.omenucache = Menu(self.optionsmenu)
        self.omenucache.add_command(label="Purge Cache",
                                    command=purgeCacheWrapper)
        self.omenucache.add_command(label="Cache Statistics",
                                    command=self.stats)
        
        self.optionsmenu.add_cascade(label="Cache", menu=self.omenucache)
        
        # Default value of 'show_cached'
        scstatus   = Options.program_options.getOption('show_cached')
        # Function to toggle the value of 'show_cached'
        sccallback = Options.program_options.makeToggleWrapper('show_cached')

        # key          = 'grab_resource_info'
        # gristatus    = Options.program_options.getOption(key)
        # gricallback  = Options.program_options.makeToggleWrapper(key)

        key          = "delete_cache_on_exit"
        dcstatus     = Options.program_options.getOption(key)
        dccallback   = Options.program_options.makeToggleWrapper(key)

        key          = "use_PIL"
        upstatus     = Options.program_options.getOption(key)
        upcallback   = Options.program_options.makeToggleWrapper(key)

        key          = "save_options_on_exit"
        sooestatus   = Options.program_options.getOption(key)
        sooecallback = Options.program_options.makeToggleWrapper(key)

        key          = "strip_carraige_returns"
        scrstatus    = Options.program_options.getOption(key)
        scrcallback  = Options.program_options.makeToggleWrapper(key)

        key          = "use_cache"
        ucstatus     = Options.program_options.getOption(key)
        uccallback   = Options.program_options.makeToggleWrapper(key)

        # This is a dictionary to hold the IntVar data that gets used to set
        # the default value of the checkbuttons.  Because of some weirdness,
        # if the data isn't held, then it does'nt work properly.  (I.e. if you
        # use a variable only within this following loop to create an IntVar
        # and don't store it outside of the temporary loop, it won't work)
        # So in other words this is data getting put into a quick dictionary
        # "Just Because" and isn't actually going to be used.
        # If you want to see what I'm talking about, try replacing
        # self.followvars with just 'var' as a temporary variable, and it
        # won't work.
        self.followvars = {}
        
        for item in [["Show Cached Status", scstatus, sccallback],
                     ["Save Options On Exit", sooestatus, sooecallback],
                     ["Strip Carriage Returns", scrstatus, scrcallback],
                     ["Use Cache", ucstatus, uccallback],
                     ["Delete Cache on Exit", dcstatus, dccallback],
                     ["Use PIL if Available", upstatus, upcallback]]:
            # Store this variable because python is weird.
            self.followvars[item[0]] = IntVar()

            if item[1]:  self.followvars[item[0]].set(1)
            else:        self.followvars[item[0]].set(0)
            
            self.optionsmenu.add_checkbutton(label=item[0],
                                             indicatoron=TRUE,
                                             # Use the stored varable here...
                                             # bind it to the menu
                                             variable=self.followvars[item[0]],
                                             command=item[2])

        try:
            if self.bookmarkTkMenu:
                self.bookmarkTkMenu = None
        except:
            pass
        
        self.bookmarkTkMenu = self.buildTkBookmarksMenu(self.menu,
                                                        self.bookmarks)
        
        self.helpmenu = Menu(self.menu)
        self.helpmenu.add_command(label='About', command=self.about)

        self.hdebug = Menu(self.helpmenu)
        
        def printopts(opts=Options.program_options):
            print opts
            
        self.hdebug.add_command(label="Dump Options", command=printopts)
        self.hdebug.add_command(label="Dump Queue", command=self.dumpQueue)
        self.helpmenu.add_cascade(label='Debug', menu=self.hdebug)        

        for menu_node in [self.filemenu, self.editmenu, self.navmenu,
                          self.bookmarkTkMenu, self.optionsmenu,
                          self.helpmenu]:
            # Disable menu tearoffs.
            menu_node['tearoff'] = FALSE

        self.menu.add_cascade(label="File", menu=self.filemenu)
        self.menu.add_cascade(label="Edit", menu=self.editmenu)
        self.menu.add_cascade(label="Navigation", menu=self.navmenu)
        self.menu.add_cascade(label="Bookmarks", menu=self.bookmarkTkMenu)
        self.menu.add_cascade(label="Options", menu=self.optionsmenu)
        self.menu.add_cascade(label="Help", menu=self.helpmenu)
# END TkGui class




# Cache.py
# $Id: Cache.py,v 1.14 2001/07/14 22:50:28 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
#
# Handles cache-file related operations.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
############################################################################
import os
import utils
import string
from gopher           import *
from Tkinter          import *
import GopherResponse
import Pmw
import Options

CacheException = "Error"

class Cache:
    verbose = None
    
    def __init__(self, *args):
        self._d = None
        return None

    def getCacheDirectory(self):
        try:
            dir = Options.program_options.getOption('cache_directory')
            return os.path.abspath(dir)
        except:
            return os.path.abspath("cache")

    def getCachePrefix(self):
        try:
            dir = Options.program_options.getOption('cache_prefix')
            return os.path.abspath(dir)
        except:
            return os.path.abspath("cache%scache" % os.sep)

    def getCacheStats(self):
        cdir = self.getCacheDirectory()
        
        [filecount, totalBytes, dircount] = utils.summarize_directory(cdir)

        kbcount = float(totalBytes)/float(1024)
        mbcount = kbcount/float(1024)

        kbcount = "%0.2fKB" % kbcount
        mbcount = "%0.2fMB" % mbcount

        datasize = "There is %s (%s) of data" % (kbcount, mbcount)
        fdcount  = "in %s files (%s directories total)" % (filecount, dircount)
        closer   = "underneath %s" % cdir

        return "%s\n%s\n%s" % (datasize, fdcount, closer)

    def emptyCache(self, parentTk):
        pref = self.getCachePrefix()
        
        self.dialog = Pmw.Dialog(parent=parentTk,
                                 title="Really delete cache?",
                                 buttons=('OK', 'Cancel'),
                                 defaultbutton='Cancel',
                                 command=self.__confirmDelete)
        
        fr = Frame(self.dialog.interior())
        fr.pack(side='top', fill='both', expand=1)

        lab1 = Label(fr, text="Are you sure you want to empty the cache?")
        lab2 = Label(fr, text="WARNING:  THIS WILL DELETE ALL FILES AND")
        lab3 = Label(fr, text="ALL SUBDIRECTORIES IN YOUR CACHE DIRECTORY,")
        lab4 = Label(fr, text="%s" % pref)
        lab5 = Label(fr, text=" ")
        lab6 = Label(fr, text="Continue?")

        lab1.grid(row=1, col=0, columnspan=5)
        lab2.grid(row=2, col=0, columnspan=5)
        lab3.grid(row=3, col=0, columnspan=5)
        lab4.grid(row=4, col=0, columnspan=5)
        lab5.grid(row=5, col=0, columnspan=5)
        lab6.grid(row=6, col=0, columnspan=5)
                                                
        # self.dialog.activate()

    def deleteCacheNoPrompt(self, *args):
        """Delete the cache without asking the user.  Do not do this unless
        you know what you're doing."""
        return self.__deleteCache()

    def __confirmDelete(self, button):
        if button != 'OK':
            self.dialog.destroy()
            return None

        self.dialog.destroy()   # Destroy the dialog anyway
        self.__deleteCache()    # Clean the files out
        return None
        
    def __deleteCache(self):
        pref = self.getCachePrefix()

        if not utils.dir_exists(pref):
            raise CacheException, "Cache prefix %s doesn't exist." % pref
                        
        cache_directories = os.listdir(pref)

        # I've been told that there's a shell utility module that does this
        # probably safer and better, but I'm not sure where it is, or whether
        # or not it's portable to crappy no-name 32 bit operating systems
        # out of Redmond, Washington.
        
        for item in cache_directories:
            item = "%s%s%s" % (pref, os.sep, item)
            if os.path.isdir(item):
                print "Recursively deleting \"%s\"" % item
                utils.recursive_delete(item)
            else:
                print "Eh?  \"%s\" isn't a directory.  That's odd..." % item

    def isInCache(self, resource):
        """Takes a resource, and returns true if the resource seems to have
        an available cache file associated with it, and None otherwise."""

        pref     = self.getCachePrefix()
        filename = resource.toCacheFilename()
        
        if pref[len(pref)-1] != os.sep and filename[0] != os.sep:
            # When joining together, separate paths with os.sep
            pref = "%s%s" % (pref, os.sep)

        # Now join together, and they will be properly joined.
        filename = pref + filename

        try:
            info = os.stat(os.path.abspath(filename))
            return [os.path.abspath(filename), info[6]]
        except OSError:
            return None
        return None
        
    def uncache(self, resource):
        """Takes a resource, and returns either None if the given resource
        is not cached on disk, or it returns a GopherResponse corresponding
        to what would have been gotten if it had been fetched from the
        server."""

        pref = self.getCachePrefix()
        file = resource.toCacheFilename()

        if pref[len(pref)-1] != os.sep and file[0] != os.sep:
            # When joining together, separate paths with os.sep
            pref = "%s%s" % (pref, os.sep)
            
        filename = pref + file
        
        try:
            # See if the file exists...
            tuple = os.stat(filename)
            if self.verbose:
                print "File %s of size %d exists." % (filename, tuple[6])
        except OSError:
            # The file doesn't exist, we can't uncache it.
            return None

        print "===> Uncaching \"%s\"" % filename
        resp = GopherResponse.GopherResponse()
        resp.setType(resource.getTypeCode())
        resp.setHost(resource.getHost())
        resp.setPort(resource.getPort())
        resp.setLocator(resource.getLocator())
        resp.setName(resource.getName())
        
        try:
            fp = open(filename, "r")

            # Consider reworking this somehow.  Slurp the entire file into
            # buffer.
            buffer = fp.read()
            fp.close()

            try:
                resp.parseResponse(buffer)
                resp.setData(None)
                # print "Loaded cache is a directory entry."
            except:
                # print "Loaded cache is not a directory."
                resp.setData(buffer)
                
            # Got it!  Loaded from cache anyway...
            # print "UNCACHE found data for use."
            return resp       
        except IOError, errstr:
            raise CacheException, "Couldn't read data on\n%s:\n%s" % (filename,
                                                                      errstr)

        # We failed.  Oh well...
        return None
    
    def cache(self, resp, resource):
        """Takes a GopherResponse and a GopherResource.  Saves the content of
        the response to disk, and returns the filename saved to."""

        if resource.isAskType():
            # Don't cache ASK blocks.  This is because if you do, the program
            # will interpret it as data, and put the question structure inside
            # a text box.   Plus, since these may be dynamic, caching could
            # lead to missing out on things.
            raise CacheException, "Do not cache AskTypes.  Not a good idea."
        
        basedir      = self.getCachePrefix()
        basefilename = resource.toCacheFilename()

        # Problem - basedir is our base directory, but basefilename contains
        # trailing filename info that shouldn't be part of the directories
        # we're creating.
        filenamecopy = basefilename[:]
        ind = string.rfind(filenamecopy, os.sep)
        
        # Chop off extra info so "/home/x/foobar" becomes "/home/x/foobar"
        # this is because make_directories will otherwise create foobar as
        # a directory when it's actually a filename
        filenamecopy = filenamecopy[0:ind]     

        # Create the directory structure where necessary
        utils.make_directories(filenamecopy, basedir)

        basedirlastchar = basedir[len(basedir)-1]
        if basedirlastchar == os.sep:
            filename = "%s%s" % (basedir, basefilename)
        else:
            filename = "%s%s%s" % (basedir, os.sep, basefilename)

        # print "Cache: caching data to \"%s\"" % filename
        
        try:
            fp = open(filename, "w")

            if resp.getData() == None:    # This is a directory entry.
                response_lines = resp.getResponses()
                # Each response line is a GopherResource
                # write them as if it was a file served by the gopher server.
                # that way it can be easily reparsed when loaded from the
                # cache.
                for response_line in response_lines:
                    fp.write(response_line.toProtocolString())

                # write the string terminator.  This isn't really needed
                # since it isn't data, but it helps fool our other objects
                # into thinking that it's dealing with data off of a socket
                # instead of data from a file.  So do it.
                fp.write("\r\n.\r\n") 
            else:
                fp.write(resp.getData())

            fp.flush()
            fp.close()
        except IOError, errstr:
            # Some error writing data to the file.  Bummer.
            raise CacheException, "Couldn't write to\n%s:\n%s" % (filename,
                                                                  errstr)
        # Successfully wrote the data - return the filename that was used
        # to save the data into.  (Absolute path)
        return os.path.abspath(filename)
        




# GopherObject.py
# $Id: GopherObject.py,v 1.17 2001/07/09 22:31:32 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
#
# Base class for GopherResource, GopherResponse
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#############################################################################

import re
import os
import utils
from string import *
from gopher import *

GopherObjectException = "Whoops...something went wrong!"

class GopherObject:
    def __init__(self,
                 typecode = None,
                 host     = None,
                 port     = None,
                 locator  = None,
                 name     = None,
                 len      = -2):
        self.__class      = "GopherObject"
        self._shouldCache = "YES"
        self.setTypeCode(typecode)
        self.setHost(host)
        self.setPort(port)
        self.setLocator(locator)
        self.setName(name)
        self.setLen(len)
        self.setDataBlock("")
        return None
    def dup(self, res):
        """Returns a copy of this object."""
        self.setLen(res.getLen())
        self.setTypeCode(res.getTypeCode())
        self.setHost(res.getHost())
        self.setLocator(res.getLocator())
        self.setName(res.getName())
        self.setPort(res.getPort())
        return self
    # Accessors, mutators
    def shouldCache(self):
        return self._shouldCache
    def setShouldCache(self, boolval):
        self._shouldCache = boolval
        return self.getShouldCache()
    def getShouldCache(self):
        return self._shouldCache
    def getLen(self):
        return self.len
    def setLen(self, newlen):
        self.len = newlen
        return self.len
    def getTypeCode(self):
        return self.type[0]
    def setDataBlock(self, block):
        self.datablock = block
        return self.datablock
    def getDataBlock(self):
        return self.datablock
    def setType(self, newtype):
        return self.setTypeCode(newtype)
    def setTypeCode(self, newtype):
        self.type = newtype
        return self.type
    def getType(self):
        """Return a string representing the type code of the response.  See
        the gopher module for more information."""
        try:
            return responses[self.type]
        except KeyError:
            return "-Unknown-"
    def getHost(self):
        return self.host
    def setHost(self, newhost):
        self.host = newhost
        return self.host
    def getPort(self):
        return self.port
    def setPort(self, newport):
        self.port = newport
        return self.port
    def getLocator(self):
        return self.locator
    def setLocator(self, newlocator):
        self.locator = newlocator
        return self.locator
    def getName(self):
        if strip(self.name) == '/':
            self.setName("%s root" % self.getHost())
        elif self.name == '' or self.name is None:
            loc = strip(self.getLocator())
            if loc == '' or loc == '/':
                self.setName("%s root" % self.getHost())
            else:
                self.setName(" ")
                # self.setName("%s %s" % (self.getHost(), self.getLocator()))
        return self.name
    def setName(self, newname):
        self.name = newname
        return self.name
    # Methods
    def toURL(self):
        """Return the URL form of this GopherResource"""
        return "gopher://%s:%s/%s%s" % (self.getHost(),
                                        self.getPort(),
                                        self.getTypeCode(),
                                        re.sub("\t", "%9;", self.getLocator()))
    def toProtocolString(self):
        """Returns the protocol string, i.e. how it would have been served
        by the server, in a string."""
        return "%s%s\t%s\t%s\t%s\r\n" % (self.getTypeCode(),
                                         self.getName(),
                                         self.getLocator(),
                                         self.getHost(),
                                         self.getPort())
    def toXML(self):
        """Returns a small XML tree corresponding to this object.  The root
        element is the name of the object.  Override me in subclasses."""
        tags = [["type",    self.type],
                ["locator", self.locator],
                ["host",    self.host],
                ["port",    self.port],
                ["name",    self.name]]

        str = "<GopherObject>"
        for tag in tags:
            str = str + "<%s>%s</%s>" % (tag[0], tag[1], tag[0])
        str = str + "</GopherObject>"

        return str
    def __str__(self):
        return self.__class
        # return self.toString()
    def toString(self):
        """Returns a string form of this object.  Mostly only good for
        debugging."""
        return ("Type: %s\nLocator: %s\nHost: %s\nPort: %s\nName: %s\n" %
                (self.getTypeCode(), self.getLocator(),
                 self.getHost(), self.getPort(), self.getName()))

    def filenameToURL(self, filename):
        """Unencodes filenames returned by toFilename() into URLs"""        
        try:
            return utils.character_replace(filename, os.sep, "/")
        except:
            # os.sep is '/' - character_replace doesn't allow replacing a
            # character with itself.  we're running Eunuchs.
            return filename

    def toCacheFilename(self):
        filename = self.toFilename()
        lastchr = filename[len(filename)-1]
        if lastchr == os.sep or lastchr == '/':
            # Pray for no name clashes...  :)
            # We have to call it something inside the leaf directory, because
            # if you don't, you get into the problem of caching responses from
            # directories as files, and not having a directory to put the items
            # in the directory in.  And you can't call the file "" on any
            # filesystem.  :)
            filename = "%sgopherdir.idx" % filename
        elif self.getTypeCode() == RESPONSE_DIR:
            filename = "%s%sgopherdir.idx" % (filename, os.sep)

        port = self.getPort()
        str_to_find = ":%d" % int(port)

        # Cut out the port portion of the filename.  This is because some
        # OS's throw up with ':' in filenames.  Bummer, but this will make it
        # hard to translate a filename/path -> URL
        ind = find(filename, str_to_find)
        if ind != -1:
            filename = "%s%s" % (filename[0:ind],
                                 filename[ind+len(str_to_find):])

        if os.sep == '\\':       # No-name mac wannabe OS's... :)
            for char in ['/', ':', ';', '%', '*', '|']:
                # This isn't necessarily a good idea, but it's somewhat
                # necessary for windows boxen.
                filename = utils.character_replace(filename, char, ' ')
            
        return filename

    def toFilename(self):
        """Returns the name of a unique file containing the elements of this
        object.  This file is not guaranteed to not exist, but it probably
        doesn't. :)  Get rid of all of the slashes, since they are
        Frowned Upon (TM) by most filesystems."""

        replaceables = ['\t', '\n', '\\', '/']
        data = self.toURL()

        if find(lstrip(lower(data)), "gopher://") == 0:
            # Chomp the "gopher://" part
            data = data[len("gopher://"):]
        
        for r in replaceables:
            try:
                data = utils.character_replace(data, r, os.sep)
            except:
                pass

        return data



#!/usr/bin/python
# Written by David Allen <mda@idatar.com>
# This is a proof of concept on how to embed the application.  Just using this
# code embeds the entire application minus a few niceties (like bookmarks, etc)
# into a popup window.
#
# This was just meant to display that the operation of the program is
# segregated from the operation of the main GUI, and how to embed the FORG
# in python/Tkinter programs.
#
# This file is released under the terms of the GNU General Public License.
##############################################################################

from Tkinter import *
import GopherResource
import forg

x = Tk()
r = GopherResource.GopherResource()
r.setURL("gopher://gopher.quux.org")
r.setName("QUUX.org")

# Create a FORG object.  You only have to tell it what your parent
# window is, and what resource it should load when it starts.
f = forg.FORG(parent_widget=x, resource=r)

# Pack it in
f.pack(fill='both', expand=1)

# Start the mini application.
x.mainloop()



# GopherResource.py
# $Id: GopherResource.py,v 1.13 2001/07/11 22:43:09 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# This class defines the information needed for a gopher resource.  That
# usually contains all the needed information about one instance of a file,
# directory, or other "thingy" on a gopher server.  This class extends
# GopherObject which gives it most of its accessors/mutators.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#############################################################################
import re
from string           import *
from urlparse         import *
from gopher           import *
import GopherConnection
import GopherObject
# import Options

class GopherResource(GopherObject.GopherObject):
    verbose   = None
    debugging = None  # Set to true for messages at the prompt, etc.
    def __init__(self,
                 type       = RESPONSE_DIR,
                 host       = "gopher.floodgap.com",
                 port       = 70,
                 locator    = "/",
                 stringName = "",
                 auxFields  = []):
        GopherObject.GopherObject.__init__(self, type, host, port,
                                           locator, stringName)
        self.__class = "GopherResource"
        if self.debugging:
            print "NEW GOPHER RESOURCE: " + self.toString()
        self.info = None
        self.setAuxFields(auxFields)
            
        return None
    
    def setInfo(self, newinfo):
        self.info = newinfo
        return self.info
    
    def getInfo(self, shouldFetch=None):
        """Returns the ResourceInformation block associated with this
        Resource.  If shouldFetch is true, the resource block will fetch
        information about itself if none is present."""
        if not self.info and shouldFetch:
            try:
                self.setInfo(conn.getInfo(self))
            except Exception, errstr:
                print "**** GopherResource couldn't get info about itself:"
                print errstr
                # This is bad.
                self.setInfo(None)

        return self.info
    
    def setAuxFields(self, fields):
        self.auxFields = fields

        if len(self.auxFields) > 0 and self.auxFields[0] == '?':
            # We need to fetch information about this one, since it's
            # going to contain ASK blocks.
            conn = GopherConnection.GopherConnection()
            
            try:
                self.setInfo(conn.getInfo(self))
            except Exception, errstr:
                print "**** GopherResource couldn't get info about itself:"
                print errstr
                # This is bad.
                self.setInfo(None)
        
        return self.auxFields
    
    def getAuxFields(self):
        return self.auxFields

    def isAskType(self):
        if not self.isGopherPlusResource():
            return None

        if len(self.auxFields) > 0 and strip(self.auxFields[0]) == '?':
            return 1
        else:
            return None
        return None
    
    def isGopherPlusResource(self):
        if len(self.auxFields) > 0:
            return 1
        else:
            return None
    
    def toProtocolString(self):
        """Overrides the GopherObject method of the same name to provide
        support for printing out the auxFields in this object."""
        return "%s%s\t%s\t%s\t%s\t%s\r\n" % (self.getTypeCode(),
                                             self.getName(),
                                             self.getLocator(),
                                             self.getHost(), self.getPort(),
                                             join(self.getAuxFields(), "\t"))
    
    def toXML(self):
        """Returns a small XML tree corresponding to this object.  The root
        element is the name of the object.  Override me in subclasses."""
        tags = [["type",    self.getType()],
                ["locator", self.getLocator()],
                ["host",    self.getHost()],
                ["port",    self.getPort()],
                ["name",    self.getName()]]

        str = "<GopherResource>"
        for tag in tags:
            str = str + "<%s>%s</%s>" % (tag[0], tag[1], tag[0])
        str = str + "</GopherResource>"

        return str

    def toURL(self):
        """Return the URL form of this GopherResource"""
        return "gopher://%s:%s/%s%s" % (self.getHost(), self.getPort(),
                                        self.getTypeCode(),
                                        self.getLocator())
    
    def setURL(self, URL):
        """Take a URL string, and convert it into a GopherResource object.
        This destructively modifies this object and returns a copy of itself

        FROM RFC 1738:
        Gopher URLs look like this:

        gopher://host:port/TypeCodeLocator

        where TypeCode is a one-character code corresponding to some entry
        in gopher.py  (hopefully) and Locator is the locator string for the
        resource.
        """
        thingys  = urlparse(URL)
        type     = thingys[0]
        hostport = thingys[1]
        resource = thingys[2]
        sublist  = split(hostport, ":", 2)
        host     = sublist[0]
        
        try:
            port     = sublist[1]
            port     = int(port)
        except IndexError:
            port     = 70
        except ValueError:
            port     = 70

        self.setHost(host)
        self.setPort(port)

        # Strip the leading slash from the locator.  
        if resource != '' and resource[0] == '/':
            resource = resource[1:]

        if len(resource) >= 2:
            newtypecode = resource[0]
            locator = resource[1:]
        else:
            newtypecode = RESPONSE_DIR
            locator = "/"
            
        self.setLocator(locator)
        self.setName(self.getLocator())
        self.setTypeCode(newtypecode)
        return self # Return a copy of me

# End GopherResource

def URLtoResource(URL):
    """Non-class method mimicing GopherResource.setURL"""
    res = GopherResource()
    return res.setURL(URL)




# Associations.py
# $Id: Associations.py,v 1.8 2001/07/11 22:43:09 s2mdalle Exp $
#
# Handles associations between file types and programs.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
########################################################################
import re
import os
from string import *

class Associations:
    AssociationsException = "foo"
    verbose = None
    DELIMITER = " => "
    
    def __init__(self):
        self.dict    = {}
        self.threads = 0
        return None
    
    def addAssociation(self, suffix, pgm):
        """Adds an association to the list.  suffix holds the file
        extension, and pgm holds the name of the executing program.  Example
        is suffix=.jpg and pgm=gimp $1"""
        
        if suffix[0] == '.':
            suffix = suffix[1:]
        self.dict[suffix] = pgm
        return None

    def save(self, filename):
        """Saves associations to filename"""
        return self.writeToFile(filename)

    def writeToFile(self, filename):
        """Writes text representation to filename that can be loaded later"""
        fp = open(filename, "w")

        lines = ["FORG associations.  This is the mechanism that allows the",
                 "FORG to launch particular programs when dealing with",
                 "different file extensions.  The format is like this:",
                 "file_suffix%sprogram_to_launch" % self.DELIMITER,
                 "where file_suffix is something such as \".jpg\", etc.",
                 "program_to_launch is the command line of the program that",
                 "you want launched.  You may use $1 to represent the file",
                 "so to launch a HTML file, you might use:",
                 ".html%s/usr/X11/bin/netscape $1" % self.DELIMITER]

        map(lambda line,f=fp: f.write("# %s\n" % line),
            lines)

        for key in self.dict.keys():
            fp.write("%s%s%s\n" % (key, self.DELIMITER, self.dict[key]))
        fp.flush()
        fp.close()
        return 1

    def loadFromFile(self, filename):
        """Loads instance of Associations from filename"""
        fp = open(filename, "r")
        self.dict = {}
        
        for line in fp.readlines():
            line = strip(line)
            if len(line) > 0 and line[0] == '#':
                continue  # Skip comments.
            
            try:
                [key, value] = split(line, self.DELIMITER, 2)
                self.addAssociation(key, value)
            except:
                print "Error parsing line in associations file: %s" % line

        fp.close()
        return 1

    def isEmpty(self):
        return len(self.dict.keys()) == 0
    
    def getFileTypes(self):
        """Returns the file suffixes the Association list knows of"""
        return self.dict.keys()
    
    def getProgramString(self, key):
        try:
            ans = self.dict[key]
            return ans
        except KeyError:
            return None
        
    def getTmpFilename(self, suffix, *args):
        return foo + "." + suffix
        return None
    
    def removeAssociation(self, suffix):
        try:
            del(self.dict[suffix])
        except KeyError:
            pass
        
    def getAssociation(self, filename):
        """Given a particular filename, find the correct association
        for it, if any."""
        # Turn it into a real filename so some programs won't go stupid
        # on us and try to look up a partial URL in a filename, particularly
        # with cache filenames.
        # print "Finding assoc for %s" % filename
        filename = "." + os.sep + filename
        matchFound = None
        ind = len(filename)-1;
        while not matchFound and ind != -1:
            str   = filename[ind+1:]
            assoc = None
            try:
                assoc = self.dict[str]
            except:
                pass
            ind = rfind(filename, ".", 0, ind)
            if assoc:
                matchFound = 1

        if ind == -1 or not matchFound:
            # print "Couldn't find association for this filetype."
            return None

        # print "Found assoc %s for filename %s" % (assoc, filename)
        return assoc
    
    def applyAssociation(self, filename, assoc=None):
        """Given a filename and an association, execute the helper program
        in order to properly process the file."""
        
        if assoc == None:
            assoc = self.getAssociation(filename)

        if assoc == None or assoc == '':
            raise AssociationsException, "No association found."
        
        # Assoc holds the program name
        assoc = re.sub("$1", "\"" + filename + "\"", assoc, 1)
        fp = os.popen(assoc)
        print "Process dump: %s" % assoc
        try:
            while 1:
                line = fp.readline()
                if line == '':
                    break;
                print line
        except:
            print "Process %s exited with an exception." % assoc
            return None
        print "Process \"%s\" finished up and exited." % assoc
        return 1

    



# List.py
# $Id: List.py,v 1.8 2001/08/12 20:40:14 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# This is a data structure similar to a doubly linked list, but with a few
# exceptions - it has to keep track of all nodes that have EVER been in the
# list, to cache old nodes for jumping around.  It also has to know that when
# you insert a completely new item, it removes everything after the current
# slot in the list.  This is because when you jump to a new document like in
# a web browser, you can't then hit the 'forward' button.
#
# The forward and back buttons in the program will correspond to methods in
# this object.
#
# Only put ListNode objects into this.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
###############################################################################

from ListNode import *

ListException = "List error"

class List:
    def __init__(self):
        self.front   = ListNode()
        self.back    = ListNode()
        self.front.setNext(self.back)
        self.front.setPrev(None)
        self.back.setPrev(self.front)
        self.back.setNext(None)
        self.current = self.front
        return None

    def goToFront(self):
        """Sets the current node to the first node in the list."""
        self.current = self.front.next
        return None

    def goToBack(self):
        """Sets the current node to the last node in the list."""
        self.current = self.back.prev
        return None
        
    def atEnd(self):
        """Predicate:  Returns true if the current item is the last item in
        the list."""
        return self.current == self.back
    
    def atFront(self):
        """Predicate:  Returns true if the current item is the first item in
        the list."""
        return self.current == self.front
    
    def itemsRemaining(self):
        """Predicate:  Returns true if the current item is not the last
        item"""
        return not self.atEnd()
    
    def isEmpty(self):
        """Predicate:  Returns true if the list contains > 0 items."""
        if self.front.getNext() == self.back:
            return 1
        else:
            return None
        
    def traverse(self, function):
        """Takes a function argument, returns a list which contains the result
        of applying the function to each item in the list successivly.  Sort
        of like map() for this list particularly."""
        n = self.front.getNext()
        listresults = [] 
        while n != self.back:
            listresults.append(function(n))
            n = n.getNext()
        return listresults

    def insertAfter(self, newnode, afterWhat):
        """Inserts newnode after afterWhat.  afterWhat may actually be either
        a ListNode in the list, or it may be the data inside a particular
        ListNode.  But it must be a reference to the *same* data, (or node)
        not a reference to equivalent data (or node)."""
        
        n = self.front.getNext()

        if newnode.__class__ != ListNode:
            raise ListException, "newnode argument must be a ListNode"

        while n != self.back.getPrev():
            if afterWhat == n or afterWhat == n.getData():
                nn = n.getNext()
                newnode.setPrev(n)
                newnode.setNext(nn)
                afterWhat.setNext(newnode)
                nn.setPrev(newnode)
                return newnode

            n = n.getNext()

        raise ListException, "cannot insert after nonexistent node."

    def removeReference(self, ref):
        """Removes ref from the list.  Note this must be a reference to
        something in the list, not just something that has the same data."""

        n = self.front.getNext()
        
        while n != self.back:
            # We're going to be very forgiving and let the user pass us a
            # reference to either the data contained in the ListNode object,
            # or a reference to the ListNode object itself.
            if n == ref or n.getData() == ref:
                np = n.getPrev()
                nn = n.getNext()
                n.setNext(None)      # Kill the links on the node we delete...
                n.setPrev(None)
                np.setNext(nn)       # Update the links on the surrounding
                nn.setPrev(np)       # nodes...
                return ref           # Get out...we only remove the 1st one.
            
            n = n.getNext()          # Next item in the list.
            
        raise ListException, "Reference not found in list"
        
    def countNodes(self, f=None):
        """Returns the number of nodes in the list."""
        if f is None:
            f = self.front
        nodes = 0
        n = f.getNext()
        while n != self.back:
            nodes = nodes + 1
            n = n.getNext()
        return nodes
    
    def prepend(self, node):
        """Inserts the given node at the front of the list."""
        self.current = self.front
        return self.insert(node, truncate=0)
    
    def postpend(self, node):
        """Inserts the given node at the very back of the list."""
        self.current = self.back.getPrev()
        return self.insert(node, 0)
    
    def insert(self, node, truncate=1):
        """Inserts node as the next item in the list.  If truncate is true,
        then all subsequent elements are dropped, and the new node becomes
        the last node in the list."""
        if truncate:
            node.setPrev(self.current)
            node.setNext(self.back)
            self.current.setNext(node)
            self.current = node
            return self.current
        else:
            oldnext = self.current.getNext()
            node.setPrev(self.current)
            node.setNext(oldnext)
            self.current.setNext(node)
            oldnext.setPrev(node)
            self.current = node
            return self.current

    def getCurrent(self):
        """Returns the current node."""
        return self.current
    
    def removeCurrent(self):
        """Removes the current node.  The current node then becomes the next
        node in the list."""
        if not self.current:
            raise ListException, "Error:  Cannot delete NONE"
        if self.current == self.front:
            raise ListException, "Cannot delete FRONT"
        if self.current == self.back:
            raise ListException, "Cannot delete BACK"
        
        one_before_this_one = self.current.getPrev()
        one_after_this_one  = self.current.getNext()
        one_before_this_one.setNext(one_after_this_one)
        one_after_this_one.setPrev(one_before_this_one)
        self.current.setPrev(None)
        self.current.setNext(None)
        self.current.setData(None)
        self.current = one_after_this_one
        return self.current

    def getFirst(self):
        """Returns the first item in the list.  Does not change the current
        node"""
        
        first = self.front.next

        if first == self.back:
            raise ListException, "The list is empty"
        return first

    def getLast(self):
        """Returns the last item in the list.  Does not change the current
        node"""
        last = self.back.prev

        if last == self.front:
            raise ListException, "The list is empty"
        return last
    
    def getNext(self):
        """Returns the next node in the list, and advances the current node."""
        next = self.current.getNext()
        
        if next == self.back or (next == None and next == self.back):
            raise ListException, "Already at the end of the list"
        elif next == None:
            raise ListException, "getNext(): Null next field"
        
        self.current = next
        return self.current
    
    def getPrev(self):
        """Returns the previous node in the list, which then becomes the
        current node."""
        prev = self.current.getPrev()
        
        if prev == self.front or (prev == None and prev == self.front):
            raise ListException, "Already at the beginning of the list"
        elif prev == None:
            raise ListException, "getPrev(): Null prev field."

        self.current = self.current.getPrev()
        return self.current

# EOF



# Bookmark.py
# $Id: Bookmark.py,v 1.14 2001/07/11 22:43:09 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
# This is a subclass of GopherResource and is used in the Bookmark
# management system of the FORG.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
############################################################################
from string import *
from Tkinter import *

import GopherResource
import ListNode
import List
import utils
import xmllib

BookmarkFactoryException = "Error"
BookmarkException        = "Error"
BookmarkMenuException    = "Error"

class Bookmark(GopherResource.GopherResource):
    def __init__(self, res=None):
        GopherResource.GopherResource.__init__(self)

        if res != None:
            # Copy data from the resource to this bookmark.
            self.setName(res.getName())
            self.setHost(res.getHost())
            self.setPort(res.getPort())
            self.setLocator(res.getLocator())
            self.setType(res.getTypeCode())

        return None
    def toXML(self, fp, indentlevel):
        """Returns an XML representation of the object.  Callers should use
        an indentlevel of 0 since this is called recursively.  fp is the file
        pointer the data should be stored in."""
        if self.verbose:
            print "Bookmark.toXML()"
        fp.write(utils.indent(indentlevel) + "<bookmark href=\"" +
                 self.getURL() + "\">\n")
        fp.write(utils.indent(indentlevel+1) + "<title>" +
                 self.getName() + "</title>\n")
        fp.write(utils.indent(indentlevel) + "</bookmark>\n")
        return 1
    def getURL(self):
        return self.toURL()
    def toData(self):
        return "%s !! gopher://%s:%s/%s" % (self.getName(),
                                            self.getHost(),
                                            self.getPort(),
                                            self.getLocator())
    def __str__(self):
        return self.toString()
    def __repr__(self):
        return self.toString()
    def toString(self):
        if self.getName() != '':
            return "%s: %s" % (self.getHost(), self.getName())
        elif self.getLocator() == '/' or self.getLocator() == '':
            return "%s Root" % self.getHost()
        else:
            return "%s:%s %s" % (self.getHost(), self.getPort(),
                                 self.getLocator())

# Not specific to the bookmark class methods...
def parseBookmark(data):
    data = strip(data)
    items = split(data, "!! ", 2)
    
    if len(items) < 2:
        print "***Bookmark parse error: Items is: %s" % join(items,"::")
        return None
    
    bmark = Bookmark()
    try:
        items[1] = strip(items[1])
        bmark.setURL(items[1])
    except Exception, estr:
        print "Returning NONE because I couldn't parse the URL: %s" % estr
        return None
    
    bmark.setName(items[0])
    return bmark

# Subclass ListNode just for naming purposes so BookmarkMenuNodes can be
# dealt with inside of BookmarkMenus
class BookmarkMenuNode(ListNode.ListNode):
    pass

# BookmarkMenu is to organize heirarchical menus of bookmarks which will be
# editable by the user.
class BookmarkMenu(List.List):
    verbose = None
    def __init__(self, menuName=" "):
        List.List.__init__(self)
        self.menuName = menuName
        return None
    def getName(self):
        if strip(self.menuName) > 0:
            return self.menuName
        else:
            self.setName("Bookmarks")
            return self.menuName
        return "Error fetching name"
    
    def setName(self, newname):
        self.menuName = newname
        return self.menuName
    
    def toString(self):
        return "BookmarkMenu: \"%s\"" % self.getName()
    
    def __str__(self):
        return self.toString()
    
    def __repr__(self):
        return self.toString()
    
    def insert(self, item, truncate=0):
        """Overrides the insert method to always pass a truncate argument of
        0 so the list is never truncated on insertions."""
        return List.List.insert(self, item, 0)
    
    def addSubmenu(self, menu):
        """Adds menu as a submenu of this menu"""
        if menu.__class__ != BookmarkMenu and menu.__class__ != Bookmark:
            raise Exception, "Cannot add a non-Bookmark/Menu as submenu"
        return self.insert(BookmarkMenuNode(menu))
    
    def toXML(self, fp, indentlevel):
        """Returns an XML representation of this object.  This is called
        recursively, so if you are calling this for the first time, use an
        indentlevel of 0.  fp holds the file pointer to write the data into"""
        
        if self.verbose:
            print "BookmarkMenu.toXML()"
            
        il = indentlevel
        i  = utils.indent
        fp.write(i(il) + "<folder>\n")
        fp.write(i(il+1) + "<title>" + self.getName() + "</title>\n")

        def fn(item, fp=fp, il=(il+1)):
            data = item.getData()
            # It doesn't matter which kind it is, whether it's a
            # Bookmark or a BookmarkMenu since they both have toXML() methods
            # and they'll take it from here.  If it's a BookmarkMenu, this
            # will happen recursively.
            data.toXML(fp, il)
            return data.getName()

        # Apply the above function to each item in the menu
        self.traverse(fn)
        
        fp.write(i(il) + "</folder> <!-- End %s -->\n" % self.getName())
        return None
    def getTkMenu(self, parent_widget, callback):
        """Return a Tk Menu object which is suitable for being added to other
        submenus.  parent_widget is the menu you will add it to, and callback
        is a function that takes exactly one argument.  That argument is the
        Bookmark object that the user-clicked item represents."""
        # This is the menu we will return.
        m = Menu(parent_widget)

        # Create a subfunction to deal with each item in the list.
        # This way since this object already extends List and we already have
        # a traverse function, we can go through the entire list in order
        def fn(listitem, cb=callback, addTo=m):
            # Each item in the list is a ListNode, not a Bookmark, so get the
            # data out of the ListNode.
            data = listitem.getData()

            def real_callback(item=data, oldCallback=cb):
                """This is a way of converting a callback that takes no
                arguments into one that takes two for our purposes.  The
                user passes in a callback function that takes only one
                parameter (the resource you want to go to) and this is
                the actual function that is bound to the action, which calls
                the user defined function with the appropriate argument"""
                return oldCallback(item)
            
            try:
                # If it's a menu, then add the submenu recursively.
                if data.__class__ == BookmarkMenu:
                    addTo.add_cascade(label=data.getName(),
                                      menu=data.getTkMenu(addTo, cb))
                else:
                    # Otherwise add a regular command into the menu.
                    addTo.add_command(label=data.getName(),
                                      command=real_callback)
            except:
                pass
            return data.getName()

        # Apply the above to each item in the list, adding menu items and
        # submenus where needed.  
        self.traverse(fn)   # Don't care about the return results.

        return m

class BookmarkFactory(xmllib.XMLParser):
    verbose = None
    
    def __init__(self):
        xmllib.XMLParser.__init__(self)
        self.menu        = None
        self.lastTag     = ''
        self.accumulator = ''
        self.currentMenu = None
        self.currentBmrk = None
        self.folders     = []
        return None
    
    def writeXML(self, fp, menu):
        """Writes an XML representation of menu to fp"""
        # Header-type stuff.
        fp.write("<?xml version=\"1.0\"?>\n")
        fp.write("<!DOCTYPE xbel SYSTEM \"xbel-1.0.dtd\">\n\n")
        fp.write("<!-- This file contains a listing of bookmarks for\n")
        fp.write("     the FORG.  Please do not edit this unless you are\n")
        fp.write("     familiar with the structure of the file. -->\n\n");
        fp.write("<xbel>\n")
        menu.toXML(fp, 1)
        fp.write("</xbel>\n")
        fp.flush()
        return 1
    
    def getMenu(self):
        """Menu object accessor"""
        return self.menu
    
    def setMenu(self, menu):
        """Menu object mutator"""
        self.menu = menu
        return self.menu
    
    def handle_data(self, data):
        """Necessary overriden method for XML parsing - when a chunk of
        data comes in, this is what is done with it."""
        self.accumulator = self.accumulator + data
        
    def syntax_error(self, message):
        """Necessary overridden method for XML parsing - handles syntax
        errors when they occur"""
        print "****Error parsing XML: ", message
        return None
    
    def unknown_starttag(self, tag, attrs):
        """Necessary overriden method for XML parsing - handles unknown
        start tags"""
        print "****Error parsing XML: Unknown tag \"%s\"" % tag
        return None
    
    def unknown_endtag(self, tag):
        """Necessary overridden method for XML parsing - handles unknown
        end tags"""
        print "****Error parsing XML: Unknown ending tag \"%s\"" % tag
        return None
    
    def unknown_charref(self, ref):
        """Necessary overridden method for XML parsing - handles unknown
        charrefs"""
        print "****Error parsing XML: uknown charref \"%s\"" % ref
        return None
    
    def unknown_entityref(self, ref):
        """Necessary overridden method for XML parsing - handles unknown
        entity references"""
        print "****Error parsing XML: unknown entityref \"%s\"" % ref
        return None
    
    def start_xbel(self, attrs):
        self.lastTag = "xbel"
        if self.verbose:
            print "<XBEL>"
        return None
    
    def start_folder(self, attrs):
        self.currentMenu = BookmarkMenu()
        self.folders.append(self.currentMenu)
        self.lastTag = "folder"
        if self.verbose:
            print "Creating new folder"
        return None
    
    def start_bookmark(self, attrs):
        self.currentBmrk = Bookmark()
        self.lastTag = "bookmark"
        
        if self.verbose:
            print "Setting URL to be ", attrs['href']
        try:
            self.currentBmrk.setURL(attrs['href'])
        except KeyError:
            print "**** Error parsing XML: bookmark is missing 'href'"
        except Exception, errstr:
            print "**** Parse error:  Couldn't parse %s: %s" % (attrs['href'],
                                                                errstr)
            self.currentBmrk = None
            return None
                
        if self.verbose:
            print "Creating new bookmark"

        return None
    def start_title(self, attrs):
        if self.lastTag == 'xbel':
            self.currentMenu.setName("Bookmarks")
        else:
            self.accumulator = ''
        return None
    def end_title(self):
        self.accumulator = strip(self.accumulator)
        
        if self.lastTag == 'xbel' or self.lastTag == 'folder':
            if self.verbose:
                print 'Setting menu name: ', self.accumulator
            self.currentMenu.setName(self.accumulator)
        elif self.lastTag == 'bookmark':
            if self.verbose:
                print "Setting bmark name: ", self.accumulator
            self.currentBmrk.setName(self.accumulator)
        else:
            if self.verbose:
                print("****Error parsing XML: Unknown lastTag %s" %
                      self.lastTag)
        self.accumulator = ''
        return None
    def end_bookmark(self):
        if self.verbose:
            print "Inserting new bmark"

        if self.currentBmrk:
            self.currentMenu.insert(BookmarkMenuNode(self.currentBmrk))
        else:
            print "**** Error parsing XML: could not insert invalid bookmark."
            
        self.lastTag = "/bookmark"
        return None
    def end_folder(self):
        try:
            finished_folder = self.folders.pop()
        except IndexError:
            print "****Error parsing XML: </folder> without <folder>"
            return None
        
        if self.verbose:
            print "Finishing up folder: %s" % finished_folder.getName()

        if len(self.folders) > 0:
            self.currentMenu = self.folders[len(self.folders)-1]

            if self.verbose:
                print("Adding submenu \"%s\" to \"%s\"" %
                      (finished_folder.getName(),
                       self.currentMenu.getName()))
            
            self.currentMenu.addSubmenu(finished_folder)
        else:
            # The stack is empty - assign the main menu to be this item
            # here.
            if self.verbose:
                print "Finished toplevel folder."
            self.menu = finished_folder

        return None
    def end_xbel(self):
        if self.verbose:
            print "/XBEL"
        
        if len(self.folders) > 0:
            print "**** Error parsing XML: premature </xbel> element."
        
    def parseResource(self, fp):
        for line in fp.read():
            self.feed(line)
        self.close()
        return self.menu




# BookmarkEditor.py
# Written by David Allen <mda@idatar.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
############################################################################

from Tkinter import *
import Pmw
import Tree
import Dialogs
import Bookmark
import os
from string import *
import GopherObject
import GopherResource
import Options

def traversal_function(node):
    if node.__class__ != Tree.Node:
        print "NODE CLASS: %s" % node.__class__
        return None
    else:
        bm = node.id
    
    if bm.__class__ == Bookmark.BookmarkMenu:
        menu = Bookmark.BookmarkMenu()
        menu.setName(bm.getName())

        # Visit each of the children.  Note that this is children as in what
        # is displayed on the screen after the user has edited the bookmarks
        # with the editor.  This is NOT the children that are listed in the
        # actual BookmarkMenu's data structure, since that may be wrong after
        # the user edits.
        for subnode in node.subnodes:
            rval = traversal_function(subnode)

            if not rval:
                print "**** That's weird.  rval ain't."
                continue

            # The items are one of two things - BookmarkMenu's or
            # BookmarkMenuNode's.  Figure out which, and add it appropriately.
            # Note that you can't insert a submenu, you have to use addSubmenu
            # which is the reason for this conditional.
            if rval.__class__ == Bookmark.BookmarkMenu:
                print "Adding submenu:  %s" % rval.getName()
                menu.addSubmenu(rval)
            else:
                # print "Adding ITEM: %s" % rval.__class__
                # print "Adding ITEM: %s %s" % (rval.__class__, rval.getName())
                menunode = Bookmark.BookmarkMenuNode(rval)
                menu.insert(menunode)

        # Return the generated menu to be added
        return menu
    
    else:  # No children...It's a BookmarkMenuNode
        return bm

class BookmarkEditor(Toplevel):
    def __init__(self, bmtree, ondestroy=None):
        Toplevel.__init__(self)
        self.title("Edit Bookmarks...")
        # Callback to be fired off when this widget is destroyed.
        self.ondestroy = ondestroy
        
        # If user tries to close the window, call self.destroy
        self.protocol("WM_DELETE_WINDOW", self.destroy)
        
        self.make_menus()
        self.config(menu=self.menu)
        self.mainBox = Frame(self)
        self.mainBox.pack(fill='both', expand=1)
        
        self.tree = Tree.Tree(self.mainBox, bmtree,
                              rootlabel="Bookmarks", lineflag=0)
        self.tree.grid(row=0, col=0, sticky='NSEW')

        # Make expandable
        self.mainBox.grid_rowconfigure(0, weight=1)
        self.mainBox.grid_columnconfigure(0, weight=1)

        self.vscrollbar = Scrollbar(self.mainBox, orient=VERTICAL)
        self.vscrollbar.grid(row=0, column=1, sticky='NS')
        self.tree.configure(yscrollcommand=self.vscrollbar.set)
        self.vscrollbar.configure(command=self.tree.yview)

        self.hscrollbar = Scrollbar(self.mainBox, orient=HORIZONTAL)
        self.hscrollbar.grid(row=1, column=0, sticky='EW')
        self.tree.configure(xscrollcommand=self.hscrollbar.set)
        self.hscrollbar.configure(command=self.tree.xview)
        
        # must get focus so keys work for demo
        self.tree.focus_set()

        # This is CRITICAL.  Make sure this is done for several reasons:  first
        # it expands the tree so the user can see the whole thing when the
        # editor pops open.  Second, unless items in the tree are expanded,
        # their data elements aren't associated with the tree branches, so in
        # order for the bookmarks to save properly, everything must have been
        # expanded at one point or another, and this just ensures that.
        self.tree.expandAll()
        
        return None

    def destroy(self, *args):
        """User closed the window.  Prompt for saving the bookmarks to
        disk."""

        print "BookmarkEditor::destroy()"
        
        def cb(buttonName, self=self):
            print "Confirm callback: ", buttonName

            if buttonName == 'OK':
                self.save()

            if self.ondestroy:
                # Call the destroy callback specified by our parent if it's
                # present.
                self.ondestroy()

            # Call superclass method to actually destroy the window.
            Toplevel.destroy(self)
            return None

        # Create a confirmation dialog box
        self._confirmDialog = Pmw.MessageDialog(self,
                                                message_text="Save Bookmarks?",
                                                buttons=('OK', 'Cancel'),
                                                defaultbutton='OK',
                                                title='Save Bookmarks?',
                                                command=cb)
        return None
    
    def getActive(self):
        i = self.tree.getActive()
        print "Active is %s class %s" % (i, i.__class__) 

    def make_menus(self, *args):
        self.menu = Menu(self)
        self.filemenu = Menu(self.menu)
        self.filemenu.add_command(label="Save",
                                  command=self.save)
        self.filemenu.add_command(label="Create Folder",
                                  command=self.createFolder)
        self.filemenu.add_command(label="Delete Folder",
                                  command=self.deleteFolder)
        self.filemenu.add_command(label="Add a Bookmark",
                                  command=self.addBookmark)
        self.filemenu.add_command(label="Close", command=self.destroy)

        self.editmenu = Menu(self.menu)
        self.editmenu.add_command(label="Cut", command=self.cut)
        self.editmenu.add_command(label="Copy", command=self.copy)
        self.editmenu.add_command(label="Paste", command=self.paste)
        self.editmenu.add_command(label="Delete", command=self.delete)

        self.testmenu = Menu(self.menu)
        self.testmenu.add_command(label="Get Active", command=self.getActive)

        self.menu.add_cascade(label="File", menu=self.filemenu)
        self.menu.add_cascade(label="Edit", menu=self.editmenu)
        self.menu.add_cascade(label="Test", menu=self.testmenu)
        return None

    def save(self, *args):
        # data_tree is a Node object, not a BookmarkMenu
        self.tree.expandAll()   # Expand all nodes so data is present in ADT's
        data_tree = self.tree.getTree()
        
        # Take the id attribute out of each Node and string them together.
        # things may have been moved around, so the links inside the data
        # structures are no good, only copy.

        bmarks = traversal_function(data_tree)
        prefs_dir = Options.program_options.getOption('prefs_directory')
        filename = prefs_dir + os.sep + "bookmarks"
        factory = Bookmark.BookmarkFactory()
        
        try:
            fp = open(filename, "w")
            factory.writeXML(fp, bmarks)
            fp.close()
        except IOError, errstr:
            e = "Could not save bookmarks to\n%s:\n%s" % (filename, errstr)
            d = Dialogs.ErrorDialog(self, e, "Error Saving Bookmarks")

    def insertBookmark(self, bookmark):
        original_cut_buffer = self.tree.getCutBuffer()
        p = self.tree.getActive().parent

        newbm = Tree.Node(parent=None,
                          name=bookmark.getName(),
                          id=bookmark,
                          closed_icon=Tree.Icons.FILE_ICON,
                          open_icon=None,
                          x=10,
                          y=10,
                          parentwidget=self.tree)

        co = Tree.Cut_Object(newbm, newbm.full_id(), None)

        # Set the cutbuffer to be the custom folder we just created.
        self.tree.setCutBuffer(co)
        # Paste it into the heirarchy
        self.tree.paste()
        # Set the cut buffer back to its original position.
        self.tree.setCutBuffer(original_cut_buffer)
        return None
    
    def addBookmark(self):
        # Prompt the user for info.  After the info is completed,
        # insertBookmark will be called with a GopherResponse object as an
        # argument.
        dialog = Dialogs.NewBookmarkDialog(parentwin=self,
                                           cmd=self.insertBookmark)
        return None
    
    def createFolder(self, folderName=None):
        # Basically, just create a Cut_Object, and then call the paste method
        # to insert it into the tree.  Make sure to preserve the cut object
        # that the tree is working with.

        if not folderName:
            self.__newFolderDialog = Dialogs.NewFolderDialog(self,
                                                             self.createFolder)
            return None
        
        original_cut_buffer = self.tree.getCutBuffer()
        bmarkmenu = Bookmark.BookmarkMenu()
        bmarkmenu.setName(folderName)

        # We have to create a Node to insert into the tree, just like all other
        # nodes in the tree.  Since we're pasting it into the heirarchy and
        # the parent is going to change, we don't need to specify one.  'id'
        # is the data associated with the Node object.
        folder_node = Tree.Node(parent=None,
                                name=folderName,
                                id=bmarkmenu,
                                closed_icon=Tree.Icons.SHUT_ICON,
                                open_icon=Tree.Icons.OPEN_ICON,
                                x=10,
                                y=10,
                                parentwidget=self.tree)

        # Create a Cut_Object.  This is done just as it is done in the
        # Node.cut() method in Tree.py - we have to use our custom created
        # node in order to create this Cut_Object.
        co = Tree.Cut_Object(folder_node, folder_node.full_id(), 1)

        # Set the cutbuffer to be the custom folder we just created.
        self.tree.setCutBuffer(co)
        
        # Paste it into the heirarchy
        self.tree.paste()
        
        # Set the cut buffer back to its original position.
        self.tree.setCutBuffer(original_cut_buffer)
        return None

    
    def deleteFolder(self, *args):
        if not self.tree.getActive().isFolder():
            errstr = "Error:\nThe selected item\nisn't a folder."
            err = Dialogs.ErrorDialog(self, errstr, "Error")
        else:
            cutBuffer = self.tree.getCutBuffer()
            # Delete the item using the cut operation
            self.tree.cut()

            # Restore the old cutbuffer.  Normally after deleting whatever we
            # deleted would be in the cut buffer, but since we're deleting and
            # not cutting, it need not be in the cutbuffer.
            self.tree.setCutBuffer(cutBuffer)
        return None

    def delete(self, *args):
        """Deletes the currently selected node out of the tree."""
        a = self.tree.getActive()

        if a == self.tree.getRoot():
            d = Dialogs.ErrorDialog(self,
                                    "Error:\nYou cannot delete the root.")
            return None

        # Get the old cut buffer
        cutbuf = self.tree.getCutBuffer()
        # Cut the item out (which overwrites the cutbuffer)
        self.cut()
        # Set the old cutbuffer back, meaning that the deleted item is gone
        # forever.
        self.tree.setCutBuffer(cutbuf)
        return None

    def cut(self, *args):
        """Cuts the selected node out of the tree."""
        a = self.tree.getActive()

        if a == self.tree.getRoot():
            # Bad mojo.  You can't cut the root node.
            d = Dialogs.ErrorDialog(self,
                                    "Error:\nYou cannot cut the root element")
            return None
        
        return self.tree.cut()
    
    def copy(self, *args):
        a = self.tree.getActive()

        if a == self.tree.getRoot():
            # Nope, can't copy the root.
            # Should we allow this?  They couldn't paste it in at the same
            # level, but they could paste it in at a sublevel.  I don't
            # know why anybody would want to have a subfolder, but they
            # might.  So let's flag this with a FIXME with the implementation
            # note that if you're going to allow this you can't use the
            # cut() method to do it.
            d = Dialogs.ErrorDialog(self,
                                    "Error:\nYou cannot copy the root element")
            return None
        
        self.cut()
        return self.paste()
    def paste(self, *args):
        if self.tree.getCutBuffer():
            return self.tree.paste()
        else:
            d = Dialogs.ErrorDialog(self,
                                    "There is no active\nbookmark to paste")








# Connection.py
# $Id: Connection.py,v 1.18 2001/09/02 17:02:23 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# Base class for socket connections.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#############################################################################
import socket
import Options
from string import *
import utils
import errno

ConnectionException = "foo"

class Connection:
    def __init__(self, *args):
        self._d = None
        self.bytes_transferred = {'sent'     : 0,
                                  'received' : 0 }
        return None
    
    def received(self, bytes):
        """Keep track of the number of bytes received on this socket object"""
        oldval = self.bytes_transferred['received']
        self.bytes_transferred['received'] = oldval + bytes
        return self.bytes_transferred['received']
    
    def sent(self, bytes):
        """Keep track of the number of bytes sent on this socket object"""
        self.bytes_transferred['sent'] = self.bytes_transferred['sent'] + bytes
        return self.bytes_transferred['sent']
    
    def getBytesSent(self):
        return self.bytes_transferred['sent']
    
    def getBytesRead(self):
        return self.bytes_transferred['received']
    
    def readloop(self, sock, bytesToRead, msgBar=None):
        """Reads bytesToRead data off of sock or until EOF if
        bytesToRead < 0.  Optionally uses msgBar to log information to the
        user."""
        timesRead = 0
        data      = ''
        CHUNKSIZE = 1024  # Default read block size.
                          # This may get overridden depending on how much data
                          # we have to read.  Optimally we want to read all of
                          # the data that's going to come in 100 steps.

        if bytesToRead < 0:
            numKBtoRead = ""   # Don't report total size since we don't know
        else:
            # Report the total size so the user has an idea of how long it's
            # going to take.
            val = float(bytesToRead) / float(1024)
            numKBtoRead = "of %0.2f kilobytes total size" % val

            if bytesToRead > (1024 * 100): # 100 Kb
                CHUNKSIZE = bytesToRead / 100

        chunk = 'foobar'
        
        while len(chunk) > 0:
            self.checkStopped(msgBar)        # Constantly make sure we should
            chunk = sock.recv(CHUNKSIZE)     
            self.received(CHUNKSIZE)
            self.checkStopped(msgBar)        # continue...
            
            timesRead = timesRead + 1
            # print "Read %s: %s" % (CHUNKSIZE, timesRead * CHUNKSIZE)
            data = data + chunk

            if bytesToRead > 0 and len(data) >= bytesToRead:
                # print "BTR=%s, len(data)=%s, breaking" % (bytesToRead,
                # len(data))

                # Disregard content length for broken gopher+ servers.
                # break
                pass
            
            if msgBar:
                # Report statistics on how far along we are...
                bytesRead = timesRead * CHUNKSIZE
                kbRead = (float(timesRead) * float(CHUNKSIZE)) / float(1024)

                if bytesToRead > 0:
                    pct = (float(bytesRead) / float(bytesToRead))*float(100)

                    # This happens sometimes when the server tells us to read
                    # fewer bytes than there are in the file. In this case,
                    # we need to only display 100% read even though it's
                    # actually more.  Becase reading 120% of a file doesn't
                    # make sense.
                    if pct >= float(100):
                        pct = float(100)
                    pctDone = ("%0.2f" % pct) + "%"
                else:
                    pctDone = ""

                msgBar.message('state',
                               "Read %d bytes (%0.2f Kb) %s %s" %
                               (bytesRead,
                                kbRead,
                                numKBtoRead,
                                pctDone))

        # Break standards-compliance because of all those terrible gopher
        # servers out there.  Return all of the data that we read, not just
        # the first bytesToRead characters of it.  This will produce the user
        # expected behavior, but it disregards content lenghts in gopher+
        if bytesToRead > 0:
            # return data[0:bytesToRead]
            return data
        else:
            return data

    def checkStopped(self, msgBar):
        """Issue a message to the user and jump out if greenlight
        isn't true."""
        if not Options.program_options.GREEN_LIGHT:
            raise ConnectionException, "Connection stopped"

    def requestToData(self, resource, request,
                      msgBar=None, grokLine=None):
        """Sends request to the host/port stored in resource, and returns
        any data returned by the server.  This may throw
        ConnectionException.  msgBar is optional.
        May throw ConnectionException if green_light ever becomes false.
        This is provided so that the user can immediately stop the connection
        if it exists."""

        utils.msg(msgBar, "Creating socket...")
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if not self.socket:
            raise GopherConnectionException, "Cannot create socket."

        self.checkStopped(msgBar)
        utils.msg(msgBar, "Looking up hostname...")
        try:
            # Try to get a cached copy of the IP address rather than
            # looking it up again which takes longer...
            ipaddr = Options.program_options.getIP(resource.getHost())
        except KeyError:
            # OK, it wasn't in the cache.  Figure out what it is,
            # and then put it in the cache.
            try:
                self.checkStopped(msgBar)
                ipaddr = socket.gethostbyname(resource.getHost())
            except socket.error, err:
                host = resource.getHost()
                estr = "Cannot lookup\n%s:\n%s" % (host, err)
                raise ConnectionException, estr

        Options.program_options.setIP(resource.getHost(), ipaddr)
        
        # At this point, ipaddr holds the actual network address of the
        # host we're contacting.
        utils.msg(msgBar,
                  "Connecting to %s:%s..." % (ipaddr, resource.getPort()))
        try:
            retval = self.socket.connect((ipaddr, int(resource.getPort())))
            #if retval != 0:
            #    errortype = errno.errorcode[retval]
            #    raise socket.error, errortype
        except socket.error, err:
            newestr = "Cannot connect to\n%s:%s:\n%s" % (resource.getHost(),
                                                         resource.getPort(),
                                                         err)
            raise ConnectionException, newestr
        
        data = ""

        self.checkStopped(msgBar)
        self.socket.send(request)    # Send full request - usually quite short
        self.checkStopped(msgBar)
        self.sent(len(request))      # We've sent this many bytes so far...

        if grokLine:   # Read the first line...this is for Gopher+ retrievals
            line = ""  # and usually tells us how many bytes to read later
            byte = ""
            
            while byte != "\n":
                self.checkStopped(msgBar)
                byte = self.socket.recv(1)
                if len(byte) <= 0:
                    print "****************BROKE out of byte loop"
                    break
                line = line + byte
    
            bytesread = len(line)
            line      = strip(line)

            try:
                if line[0] == '+':
                    bytecount = int(line[1:]) # Skip first char: "+-1" => "-1"
                    resource.setLen(bytecount)
                    
                    # Read all of the data into the 'data' variable.
                    data = self.readloop(self.socket, bytecount, msgBar)
                else:
                    data = self.readloop(self.socket, -1, msgBar)
            except:
                print "*** Couldn't read bytecount: skipping."
                data = self.readloop(self.socket, -1, msgBar)
        else:
            data = self.readloop(self.socket, -1, msgBar)

        utils.msg(msgBar, "Closing socket.")
        self.socket.close()

        # FIXME: 'data' may be huge.  Buffering?  Write to cache file here
        # and return a cache file name?
        return data



# GUIAskForm.py
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
#
# The Tk/widget incarnation of AskForm
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
########################################################################

from Tkinter import *
from gopher import *
import Pmw
import ContentFrame
import GopherResource
import GopherResponse
import GopherConnection
import AskForm
import Question
import GUIQuestion

class GUIAskForm(ContentFrame.ContentFrame, Frame):
    verbose = None
    
    def __init__(self, parent_widget, parent_object, resp,
                 resource, filename=None, menuAssocs={}):
        Frame.__init__(self, parent_widget)  # Superclass constructor
        self.resource = resource
        self.response = resp
        self.parent   = parent_object
        self.filename = filename

        self.question_widgets = []
        self.questionBox = Pmw.ScrolledFrame(self,
                                             horizflex='fixed',
                                             vertflex ='fixed',
                                             hscrollmode='dynamic',
                                             vscrollmode='dynamic')
        return None
    
    def pack_content(self, *args):
        for x in range(0, self.response.questionCount()):
            q = self.response.nthQuestion(x)
            try:
                typename = questions_types["%s" % q.getType()]
            except KeyError:
                typename = "(ERROR NO TYPE)"
            print "PROCESSING Question %s %s %s" % (typename,
                                                    q.getDefault(),
                                                    q.getPromptString())
            try:
                wid = GUIQuestion.GUIQuestion(self.questionBox.interior(), q)
            except Exception, errstr:
                print "Couldn't make wid: %s" % errstr
                continue

            wid.grid(row=x, col=0, sticky=W)
            self.question_widgets.append(wid)

        self.submit = Button(self,
                             text="Submit", command=self.submit)
        self.questionBox.pack(side='top', expand=1, fill='both')
        self.submit.pack(side='bottom')
        return None
    
    def find(self, term, caseSensitive=None, lastIdentifier=None):
        self.parent.genericMessage("Sorry, AskForms are not searchable.\n" +
                                   "Or at least, not yet.")
        return None
    
    def submit(self, *args):
        """Submit the answers to the form questions to the server for
        processing, and load the results page."""
        values = []
        retstr = ""
        for widget in self.question_widgets:
            if widget.getType() == QUESTION_NOTE:
                continue
            values.append(widget.getResponse())
            retstr = "%s%s" % (retstr, widget.getResponse())

        print "Retstr is:\n%s" % retstr
        getMe = GopherResource.GopherResource()

        # Shouldn't cache this resource.
        getMe.setShouldCache(None)
        getMe.dup(self.resource)    # Copy defaults

        l = len(retstr)

        # Tell the server how much data we're sending...
        retstr = "+%d\r\n%s" % (l, retstr)
        
        getMe.setDataBlock(retstr)
        return self.parent.goElsewhere(getMe)



# ListNode.py
# $Id: ListNode.py,v 1.6 2001/07/11 22:43:09 s2mdalle Exp $
# Nodes to be placed in List objects.
# Subclass freely!  Life is short!
#
# Written by David Allen <mda@idatar.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#############################################################################

class ListNode:
    def __init__(self, data=None, next=None, prev=None):
        self.next = next
        self.prev = prev
        self.data = data
        return None
    
    def __str__(self):
        return self.data.__str__
    
    def __repr__(self):
        return self.data.__repr__
    
    def getNext(self):
        """Return the next item in the list"""
        return self.next
    
    def getPrev(self):
        """Return the previous item in the list"""
        return self.prev
    
    def getData(self):
        """Return the data inside this object Node"""
        return self.data
    
    def setNext(self, newnext):
        """Set the next item in the list"""
        self.next = newnext
        return self.next
    
    def setPrev(self, newprev):
        """Set the previous item in the list"""
        self.prev = newprev
        return self.prev
    
    def setData(self, data):
        """Set the data inside the object."""
        self.data = data
        return self.data




# GUISaveFile.py
# $Id: GUISaveFile.py,v 1.15 2001/07/05 17:16:53 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# This class asks users graphically which filename they'd like to save certain
# files into.  This is generally used for downloaded files that don't have
# associations bound to them, and that also can't be displayed, because
# they're binary data or otherwise ugly to look at as text.  :)  (e.g.
# *.zip, *.tar.gz, *.tgz, binhex, etc.
#
# Since this is a last resort, this module will try to import the Python
# Imaging Library (PIL) to display images if it is present.  Note that this
# requires that there be no association for the image files, since if there
# is, it won't even get as far as creating a GUISaveFile object to display.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##############################################################################

from Tkinter import *
from gopher import *
from string import *
import os
import tkFileDialog
import Pmw
import re
import Options

import ContentFrame
import GopherResource
import GopherResponse
import Dialogs

try:
    import PIL
    import Image
    import ImageTk
except:
    print "Bummer dude!  You don't have the PIL installed on your machine!"
    print "That means that the \"Use PIL\" option is going to be irrelevant"
    print "for you."

class PILImage(Label):
    # This was borrowed and adapted from the PIL programming examples.
    def __init__(self, master, im):
        if im.mode == "1":
            # bitmap image
            self.image = ImageTk.BitmapImage(im, foreground="white")
            Label.__init__(self, master, image=self.image, bg="black", bd=0)
        else:
            # Photo image object would be better...
            self.image = ImageTk.PhotoImage(im)
            Label.__init__(self, master, image=self.image, bd=0)

class GUISaveFile(ContentFrame.ContentFrame, Frame):
    verbose = None
    
    def __init__(self, parent_widget, parent_object, resp,
                 resource, filename):
        Frame.__init__(self, parent_widget)  # Superclass constructor
        self.r1 = None
        self.r2 = None
        self.filename = filename[:]
        self.parent   = parent_object
        self.response = resp
        self.resource = resource

        # Don't even try to use the PIL unless the option is set
        # to allow it.
        usePIL = Options.program_options.getOption("use_PIL")

        if usePIL and self.canDisplay():
            try:
                self.packImageContent()
            except Exception, errstr:
                self.packSaveContent()
        else:
            self.packSaveContent()
            print "Packed save content"
        return None

    def packImageContent(self, *args):
        self.createPopup()
        self.image = Image.open(self.filename)
        self.scrframe = Pmw.ScrolledFrame(self)
        imgwidget = PILImage(self.scrframe.interior(), self.image)
        imgwidget.pack()
        imgwidget.bind('<Button-3>', self.popupMenu)
        self.scrframe.pack(fill='both', expand=1)
        
        # Return and DON'T display the save file as box only if the
        # file had no problems.  Otherwise an exception was raised.
        return None

    def revertToSaveDialog(self, *args):
        self.scrframe.pack_forget()     # Unpack the scrolled frame
        self.scrframe = None            # Lose the ref to the scrolled frame
        self.image    = None            # Lose the ref to the image object.
        self.packSaveContent()          # Go back to the "Save File" content
        self.pack_content()
        return None

    def imageInfo(self, *args):
        try:
            if not self.image:
                return None
        except:
            return None
        
        info = "Bands: %s" % join(self.image.getbands(), ", ")
        size = self.image.size
        info = "%s\nWidth: %d pixels\nHeight: %d pixels" % (info,
                                                            size[0], size[1])
        info = "%s\nMode: %s" % (info, self.image.mode)

        for key in self.image.info.keys():
            info = "%s\n%s = %s" % (info, key, self.image.info[key])
        
            d = Dialogs.ErrorDialog(self, errstr=info,
                                    title='Image Information (PIL)')
        return None

    def createPopup(self):
        """Pop-up menu on right click on a message"""
        self.popup = Menu(self)
        self.popup['tearoff'] = FALSE
        self.popup.add_command(label='Save',
                               command=self.revertToSaveDialog)
        self.popup.add_command(label='Info',
                               command=self.imageInfo)

    def popupMenu(self, event):
        """Display pop-up menu on right click on a message"""
        self.popup.tk_popup(event.x_root, event.y_root)

    def canDisplay(self):
        try:
            fn = Image.open
        except:
            return None
            
        s = self.filename
        ind = rfind(s, ".")
        
        if ind != -1 and ind != (len(s)-1):
            fileExtension = lower(s[ind:])

        return 1
            
    def find(self, term, caseSensitive=None, lastIdentifier=None):
        self.parent.genericError("Error: Save Dialogs\n" +
                                 "are not searchable.")
        return None

    def pack_content(self, *args):
        return None
    
    def packSaveContent(self, *args):
        # Explicit copy - damn python and its ability to trip me up with all
        # that "everything's a reference" stuff.  :)
        default_filename = self.filename[:]

        for char in ['/', ':', ' ', '\\']:
            strtofind = "%" + "%d;" % ord(char)
            default_filename = re.sub(strtofind, char, default_filename)

        for separator in ['/', ':', '\\']:
            ind = rfind(default_filename, separator)
            if ind != -1:
                default_filename = default_filename[ind+len(separator):]
                break

        if self.useStatusLabels:
            labeltext = "%s:%d" % (resource.getHost(), int(resource.getPort()))
        
            if resource.getName() != '' and resource.getLocator() != '':
                label2 = "\"%s\" ID %s" % (resource.getName(),
                                           resource.getLocator())
            else:
                label2 = "    "

            if len(label2) > 50:
                label2 = label2[0:47] + "..."

            Label(self, text=labeltext).pack(expand=0, fill='x')
            Label(self, text=label2).pack(expand=0, fill='x')

            Label(self, text=" ").pack()  # Empty line.

        Label(self,
              text="Please enter a filename to save this file as:").pack()
        
        cframe = Frame(self)
        cframe.pack(expand=1, fill='both')
        Label(cframe, text="Filename:").pack(side='left')

        self.filenameEntry = Entry(cframe)
        self.filenameEntry.insert('end', default_filename)
        self.filenameEntry.pack(side='left', expand=1, fill='x')
        self.filenameEntry.bind("<Return>", self.save)
        Button(cframe, text="Browse", command=self.browse).pack(side='right')
        self.saveButton = Button(cframe, text="Save", command=self.save)
        self.saveButton.pack(side='right')

        return None
    
    def browse(self, *args):
        dir = os.path.abspath(os.getcwd())
        filename = tkFileDialog.asksaveasfilename(initialdir=dir)

        if filename:
            self.filenameEntry.delete(0, 'end')
            self.filenameEntry.insert('end', filename)

        return None
    def save(self, *args):
        filename = self.filenameEntry.get()

        try:
            fp = open(filename, "w")
            fp.write(self.response.getData())
            fp.flush()
            fp.close()
        except IOError, errstr:
            self.parent.genericError("Couldn't save file\n%s:\n%s" % (filename,
                                                                      errstr))
            return None

        if self.r1:
            self.r1.destroy()
        if self.r2:
            self.r2.destroy()
            
        self.r1 = Label(self, text="File successfully saved into").pack()
        self.r2 = Label(self, text=filename).pack()
        return None



# GUIError.py
# $Id: GUIError.py,v 1.3 2001/02/22 03:01:03 s2mdalle Exp $
#
# Displays errors similar to "Cannot Load Page" in the main window.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##########################################################################

from gopher import *
from Tkinter import *
import Pmw
from string import *

import GopherResource
import GopherResponse
import GopherConnection
import ContentFrame
import Cache
import Options

class GUIError(ContentFrame.ContentFrame, Frame):
    verbose = None

    def __init__(self, parent_widget, resource, error_message):
        Frame.__init__(self, parent_widget)
        ContentFrame.ContentFrame.__init__(self)
        self.parent = parent_widget

        Label(self, text="Unable to load").pack()
        Label(self, text=resource.toURL()).pack()
        Label(self, text=error_message).pack()
    def pack_content(self, *args):
        return None
    def find(self, term, caseSensitive=None, lastIdentifier=None):
        self.parent.genericError("Error:  Error messages\n" +
                                 "are not searchable.")
        return None



# AskForm.py
# Written by David Allen <mda@idatar.com>
# Released under the terms of the GNU General Public License
# $Id: AskForm.py,v 1.6 2001/04/07 19:12:44 s2mdalle Exp $
# An AskForm is essentially a conglomeration of Question objects.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#########################################################################

from string import *
import GopherResponse
import Question

class AskForm(GopherResponse.GopherResponse):
    def __init__(self, askformdata=""):
        GopherResponse.GopherResponse.__init__(self)
        self.questions = []
        self.setAskFormData(askformdata)
        return None
    def questionCount(self):
        return len(self.questions)
    def nthQuestion(self, nth):
        return self.questions[nth]
    def setAskFormData(self, data):
        self.data = data

        print "ASKFORM:  Parsing data block:\n", data
        self.lines = split(self.data, "\n")

        for line in self.lines:
            line = strip(line)
            if line == '' or line == '.':
                continue
            try:
                q = Question.Question(line)
            except Question.QuestionException, qstr:
                print "Error parsing question \"%s\": %s" % (line, qstr)
                continue
            
            self.questions.append(q)



# ContentFrame.py
# $Id: ContentFrame.py,v 1.6 2001/03/27 22:18:02 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# This is the base class for anything that gets displayed by the program
# in the main window to represent a response from the server.  In other words,
# things such as text boxes and directory listings from gopher servers should
# extend this class.
#
# This class defines the behavior of anything that is presented to the user.
# OK, so I'm a bit new to python.  What I mean here is like a Java interface.
# I know I can't force subclasses to 'implement' this function, but they
# should, otherwise the functions here will spit out obscene messages to the
# user.  :)
#
# Note that creating the object is NOT ENOUGH.  You must call pack_content
# after creating it to actually add all the things that the widget will display
# this is a workaround for some odd behavior in Pmw.  For this reason you
# should NOT rely on nothing being present in the widget if you don't call
# pack_content.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
###############################################################################

from Tkinter import *
import Pmw

class ContentFrame:
    useStatusLabels = None
    
    def __init__(self):
        pass
    def pack_content(self, *args):
        """Packs the content of the box into the frame.  Note this does NOT
        pack this object into its parent, only its children into itself."""
        print "ContentFrame.pack_content:  Superclass failed to override."
        return None
    def find(self, term, caseSensitive=None, lastIdentifier=None):
        """Find some term within a frame.  If caseSensitive is true, then the
        search will be caseSensitive.  lastIdentifier is something that was
        previously returned by this function as the last match.  If you want
        to find the next occurance of something, pass the last occurance in
        and it will search from then on."""
        print "**********************************************************"
        print "***** ContentFrame.find():  Some subclass fucked up. *****"
        print "**********************************************************"
        return None



# gopher.py
# $Id: gopher.py,v 1.7 2001/07/11 22:43:09 s2mdalle Exp $
# Gopher protocol definitions.
# Written by David Allen <mda@idatar.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
########################################################################
RESPONSE_FILE    = '0' # Item is a file
RESPONSE_DIR     = '1' # Item is a directory
RESPONSE_CSO     = '2' # Item is a CSO phone-book server
RESPONSE_ERR     = '3' # Error
RESPONSE_BINHEX  = '4' # Item is a BinHexed Macintosh file.
RESPONSE_DOSBIN  = '5' # Item is DOS binary archive of some sort.
RESPONSE_UUE     = '6' # Item is a UNIX uuencoded file.
RESPONSE_INDEXS  = '7' # Item is an Index-Search server.
RESPONSE_TELNET  = '8' # Item points to a text-based telnet session.
RESPONSE_BINFILE = '9' # Item is a binary file!
RESPONSE_REDSERV = '+' # Item is a redundant server
RESPONSE_TN3270  = 'T' # Item points to a text-based tn3270 session.
RESPONSE_GIF     = 'g' # Item is a GIF format graphics file.
RESPONSE_IMAGE   = 'I' # Item is some kind of image file.
RESPONSE_UNKNOWN = '?' # Unknown.  WTF?
RESPONSE_BITMAP  = ':' # Gopher+ Bitmap response
RESPONSE_MOVIE   = ";" # Gopher+ Movie response
RESPONSE_SOUND   = "<" # Gopher+ Sound response

# The following are types not found in the RFC definition of gopher but that
# I've encountered on the net, so I added
RESPONSE_BLURB   = 'i'
RESPONSE_HTML    = 'h'

# Gopher+ errors
ERROR_NA         = '1' # Item is not available.
ERROR_TA         = '2' # Try again later (eg. My load is too high right now.)
ERROR_MOVED      = '3' # Item has moved.

responses = { "%s" % RESPONSE_FILE    : "File:",
              "%s" % RESPONSE_DIR     : "Directory:",
              "%s" % RESPONSE_CSO     : "CSO phone-book server:",
              "%s" % RESPONSE_ERR     : "Error:",
              "%s" % RESPONSE_BINHEX  : "BinHexed Macintosh file:",
              "%s" % RESPONSE_DOSBIN  : "DOS binary archive:",
              "%s" % RESPONSE_UUE     : "UNIX UUEncoded file:",
              "%s" % RESPONSE_INDEXS  : "Index-Search server:",
              "%s" % RESPONSE_TELNET  : "Telnet session:",
              "%s" % RESPONSE_BINFILE : "Binary file:",
              "%s" % RESPONSE_REDSERV : "Redundant server:",
              "%s" % RESPONSE_TN3270  : "tn3270 session:",
              "%s" % RESPONSE_GIF     : "GIF file:",
              "%s" % RESPONSE_IMAGE   : "Image file:",
              "%s" % RESPONSE_BLURB   : " ",
              "%s" % RESPONSE_HTML    : "HTML file:",
              "%s" % RESPONSE_BITMAP  : "Bitmap Image:",
              "%s" % RESPONSE_MOVIE   : "Movie:",
              "%s" % RESPONSE_SOUND   : "Sound:",
              "%s" % RESPONSE_UNKNOWN : "Unknown:" }
errors    = { "%s" % ERROR_NA         : "Error: Item is not available.",
              "%s" % ERROR_TA         : "Error: Try again (server busy)",
              "%s" % ERROR_MOVED      : "Error: This resource has moved." }

# There is nothing special about these numbers, just make sure they're
# all unique.
QUESTION_ASK       = 20
QUESTION_ASKP      = 21
QUESTION_ASKL      = 22
QUESTION_ASKF      = 23
QUESTION_SELECT    = 24
QUESTION_CHOOSE    = 25
QUESTION_CHOOSEF   = 26
QUESTION_NOTE      = 27

# Mapping from Gopher+ types to internal values.
questions = { "Ask"             : QUESTION_ASK,
              "AskP"            : QUESTION_ASKP,
              "AskL"            : QUESTION_ASKL,
              "AskF"            : QUESTION_ASKF,
              "Select"          : QUESTION_SELECT,
              "Choose"          : QUESTION_CHOOSE,
              "ChooseF"         : QUESTION_CHOOSEF,
              "Note"            : QUESTION_NOTE }

questions_types = { "%s" % QUESTION_ASK          : "Ask",
                    "%s" % QUESTION_ASKP         : "AskP",
                    "%s" % QUESTION_ASKL         : "AskL",
                    "%s" % QUESTION_ASKF         : "AskF",
                    "%s" % QUESTION_SELECT       : "Select",
                    "%s" % QUESTION_CHOOSE       : "Choose",
                    "%s" % QUESTION_CHOOSEF      : "ChooseF" }

# Colors - not related to gopher protocol functioning, but useful.
RED         = "#FF0000"
GREEN       = "#00FF00"
BLUE        = "#0000FF"
WHITE       = "#FFFFFF"
BLACK       = "#000000"



# GUIFile.py
# $Id: GUIFile.py,v 1.12 2001/07/05 17:16:53 s2mdalle Exp $
# Written by David Allen <mda@idatar.com>
#
# This is the class that describes how files behave when loaded into
# the FORG.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##############################################################################

from Tkinter import *
from gopher import *
from string import *
import Pmw
import re
import Options

import ContentFrame
import GopherResource
import GopherResponse

class GUIFile(ContentFrame.ContentFrame, Frame):
    verbose = None
    
    def __init__(self, parent_widget, parent_object, resp,
                 resource, filename):
        Frame.__init__(self, parent_widget)  # Superclass constructor
        self.resp = resp

        if self.useStatusLabels:
            labeltext = "%s:%d" % (resource.getHost(), int(resource.getPort()))

            if resource.getName() != '' and resource.getLocator() != '':
                label2 = "\"%s\" ID %s" % (resource.getName(),
                                           resource.getLocator())
            else:
                label2 = "    "

            if len(label2) > 50:
                label2 = label2[0:47] + "..."

            Label(self, text=labeltext).pack(expand=0, fill='x')
            Label(self, text=label2).pack(expand=0, fill='x')

        if resp.getTypeCode() != RESPONSE_FILE:
            Label(self, text="This file has been saved in:").pack()
            Label(self, text=filename).pack()
        else:
            self.textwid = Pmw.ScrolledText(self, hscrollmode='dynamic',
                                            vscrollmode='static')
            tw = self.textwid.component('text')
            tw.configure(background='#FFFFFF', foreground='#000000')
                
            self.textwid.component('text').bind('<Button-3>',
                                                parent_object.popupMenu)
            
            self.textwid.pack(expand=1, fill='both')
        return None

    def pack_content(self, *args):
        if Options.program_options.getOption('strip_carraige_returns'):
            # print "Stripping carriage returns..."
            data = replace(self.resp.getData(), "\r", "")
        else:
            data = self.resp.getData()

        if len(data) < 1024:
            self.textwid.settext(data)
        else:
            for index in range(0, len(data), 500):
                self.textwid.insert('end', data[index:index+500])
        return None
    
    def destroy(self, *args):
        self.pack_forget()
        self.textwid.destroy()
        Frame.destroy(self)
        return None

    def find(self, term, caseSensitive=None, lastIdentifier=None):
        """Overrides the function of the same type from ContentFrame"""
        try:
            # This will raise an exception if it's a 'save' type layout
            # where the data isn't displayed to the user.
            tw = self.textwid.component('text')
            print "Component is ", tw
        except:
            # Don't mess with this.  The user can read the entire label, all
            # big bad few lines of it.
            raise Exception, "This window is not searchable."

        if lastIdentifier is None:
            lastIdentifier = '0.0'

        # The variable caseSensitive is true if the search is case sensitive,
        # and false otherwise.  But since tkinter wants to know whether or not
        # the search is case INsensitive, we flip the boolean value, and use
        # it for the 'nocase' keyword arg to the search method.
        csflipped = (not caseSensitive)
        pos = tw.search(pattern=term, forwards=1,
                        nocase=csflipped, index=lastIdentifier,
                        stopindex=END)
        
        if pos:
            # Find the real index of the position returned.
            found_index = tw.index(pos)
        else:
            found_index = None

        print "Found index is \"%s\"" % found_index

        if found_index:
            tw.yview(found_index)
            ending_index = tw.index("%s + %d chars" % (found_index, len(term)))
            # Set the selection to highlight the given word.
            tw.tag_add(SEL, found_index, ending_index)
            return tw.index("%s + 1 chars" % found_index)

        return None



# utils.py
# Written by David Allen <mda@idatar.com>
# $Id: utils.py,v 1.12 2001/07/05 17:16:53 s2mdalle Exp $
# Random functions used in many places that don't belong together in an
# object.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#############################################################################
from string import *
import os
import stat

def summarize_directory(dirname):
    """Returns an array [x, y, z] where x is the number of files in all subdirs
    of dirname and y is the total summed size of all of the files and z is
    the total number of directories."""

    filecount  = 0
    summedSize = 0
    dircount   = 0
    
    files = os.listdir(dirname)
    for file in files:
        if file == '.' or file == '..':
            continue
        
        path = os.path.join(dirname, file)
        if os.path.isdir(path):
            dircount = dircount + 1     # Discovered a new directory
            [tfc, tfs, tdc] = summarize_directory(path)
            # Add whatever came back from the recursive call.
            filecount  = filecount + tfc
            summedSize = summedSize + tfs
            dircount   = dircount + tdc
        else:
            filecount = filecount + 1
            statinfo  = os.stat(path)
            summedSize = summedSize + statinfo[stat.ST_SIZE]

    return [filecount, summedSize, dircount]

def file_exists(filename):
    try:
        os.stat(filename)
    except OSError:
        return None

    return 1

def recursive_delete(dirname):
    """Runs the equivalent of an rm -rf on a given directory.  Does not make
    any distinction between symlinks, etc. so use with extreme care.
    Thanks to comp.lang.python for tips on this one..."""
    files = os.listdir(dirname)
    for file in files:
        if file == '.' or file == '..':
            # The calls used to find filenames shouldn't ever return this,
            # but just in case, we check since this would be horrendously
            # bad.
            continue
        
        path = os.path.join(dirname, file)
        if os.path.isdir(path):
            recursive_delete(path)
        else:
            print 'Removing file: "%s"' % path
            retval = os.unlink(path)

    print 'Removing directory:', dirname
    os.rmdir(dirname)
    return 1

def character_replace(str, findchar, replacechar):

    if findchar is replacechar or findchar == replacechar:
        # That's a no-no...
        raise Exception, "character_replace: findchar == replacechar"

    ind = find(str, findchar)

    while ind != -1:
        str = str[0:ind] + "%s" % replacechar + str[ind+1:]
        ind = find(str, findchar)
    return str

def dir_exists(dirname):
    try:
        stat_tuple = os.stat(dirname)
    except OSError:
        return None
    return stat.S_ISDIR(stat_tuple[0])

def make_directories(path, basedir):
    """Makes path directories off of basedir.  Path is a relative path,
    and basedir is an absolute path.  Example of invocation:
    make_directories('foo/bar/baz/quux', '/home/user/') will ensure that
    the path /home/user/foo/bar/baz/quux exists"""
    arr = split(path, os.sep)

    if basedir[len(basedir)-1] == os.sep:
        # Trim tailing dir separator
        basedir = basedir[0:len(basedir)-1]

    if not dir_exists(basedir):
        os.mkdir(basedir)

    for item in arr:
        if not item or item == '':
            continue
        dirname = "%s%s%s" % (basedir, os.sep, item)

        if not dir_exists(dirname):
            os.mkdir(dirname)
            
        basedir = dirname
    return 1

def msg(msgBar, msgtxt):
    """Puts msgtext into the msgBar.  Does nothing if msgBar is None"""
    if msgBar != None:
        msgBar.message('state', msgtxt)
    #else:
    #    print "=> msgbar: %s" % msgtxt

def indent(indentlevel=1):
    str = ""
    if indentlevel < 0:
        raise Exception, "Indentlevel < 0 - you can't do that!  :)"
    while indentlevel > 0:
        str = str + "  "
        indentlevel = indentlevel - 1
    return str
